import contextlib
import importlib
import os
import platform
import re
import shutil
import site
import sys
import tempfile
import textwrap
from distutils import sysconfig
from distutils.command.build_ext import build_ext
from distutils.core import Distribution
from distutils.errors import (
    CompileError,
    DistutilsPlatformError,
    DistutilsSetupError,
    UnknownFileError,
)
from distutils.extension import Extension
from distutils.tests import missing_compiler_executable
from distutils.tests.support import (
    TempdirManager,
    copy_xxmodule_c,
    fixup_build_ext,
)
from io import StringIO

import jaraco.path
import path
import pytest
from test import support

from .compat import py38 as import_helper


@pytest.fixture()
def user_site_dir(request):
    self = request.instance
    self.tmp_dir = self.mkdtemp()
    self.tmp_path = path.Path(self.tmp_dir)
    from distutils.command import build_ext

    orig_user_base = site.USER_BASE

    site.USER_BASE = self.mkdtemp()
    build_ext.USER_BASE = site.USER_BASE

    # bpo-30132: On Windows, a .pdb file may be created in the current
    # working directory. Create a temporary working directory to cleanup
    # everything at the end of the test.
    with self.tmp_path:
        yield

    site.USER_BASE = orig_user_base
    build_ext.USER_BASE = orig_user_base


@contextlib.contextmanager
def safe_extension_import(name, path):
    with import_helper.CleanImport(name):
        with extension_redirect(name, path) as new_path:
            with import_helper.DirsOnSysPath(new_path):
                yield


@contextlib.contextmanager
def extension_redirect(mod, path):
    """
    Tests will fail to tear down an extension module if it's been imported.

    Before importing, copy the file to a temporary directory that won't
    be cleaned up. Yield the new path.
    """
    if platform.system() != "Windows" and sys.platform != "cygwin":
        yield path
        return
    with import_helper.DirsOnSysPath(path):
        spec = importlib.util.find_spec(mod)
    filename = os.path.basename(spec.origin)
    trash_dir = tempfile.mkdtemp(prefix='deleteme')
    dest = os.path.join(trash_dir, os.path.basename(filename))
    shutil.copy(spec.origin, dest)
    yield trash_dir
    # TODO: can the file be scheduled for deletion?


@pytest.mark.usefixtures('user_site_dir')
class TestBuildExt(TempdirManager):
    def build_ext(self, *args, **kwargs):
        return build_ext(*args, **kwargs)

    def test_build_ext(self):
        missing_compiler_executable()
        copy_xxmodule_c(self.tmp_dir)
        xx_c = os.path.join(self.tmp_dir, 'xxmodule.c')
        xx_ext = Extension('xx', [xx_c])
        dist = Distribution({'name': 'xx', 'ext_modules': [xx_ext]})
        dist.package_dir = self.tmp_dir
        cmd = self.build_ext(dist)
        fixup_build_ext(cmd)
        cmd.build_lib = self.tmp_dir
        cmd.build_temp = self.tmp_dir

        old_stdout = sys.stdout
        if not support.verbose:
            # silence compiler output
            sys.stdout = StringIO()
        try:
            cmd.ensure_finalized()
            cmd.run()
        finally:
            sys.stdout = old_stdout

        with safe_extension_import('xx', self.tmp_dir):
            self._test_xx()

    @staticmethod
    def _test_xx():
        import xx

        for attr in ('error', 'foo', 'new', 'roj'):
            assert hasattr(xx, attr)

        assert xx.foo(2, 5) == 7
        assert xx.foo(13, 15) == 28
        assert xx.new().demo() is None
        if support.HAVE_DOCSTRINGS:
            doc = 'This is a template module just for instruction.'
            assert xx.__doc__ == doc
        assert isinstance(xx.Null(), xx.Null)
        assert isinstance(xx.Str(), xx.Str)

    def test_solaris_enable_shared(self):
        dist = Distribution({'name': 'xx'})
        cmd = self.build_ext(dist)
        old = sys.platform

        sys.platform = 'sunos'  # fooling finalize_options
        from distutils.sysconfig import _config_vars

        old_var = _config_vars.get('Py_ENABLE_SHARED')
        _config_vars['Py_ENABLE_SHARED'] = True
        try:
            cmd.ensure_finalized()
        finally:
            sys.platform = old
            if old_var is None:
                del _config_vars['Py_ENABLE_SHARED']
            else:
                _config_vars['Py_ENABLE_SHARED'] = old_var

        # make sure we get some library dirs under solaris
        assert len(cmd.library_dirs) > 0

    def test_user_site(self):
        import site

        dist = Distribution({'name': 'xx'})
        cmd = self.build_ext(dist)

        # making sure the user option is there
        options = [name for name, short, label in cmd.user_options]
        assert 'user' in options

        # setting a value
        cmd.user = True

        # setting user based lib and include
        lib = os.path.join(site.USER_BASE, 'lib')
        incl = os.path.join(site.USER_BASE, 'include')
        os.mkdir(lib)
        os.mkdir(incl)

        # let's run finalize
        cmd.ensure_finalized()

        # see if include_dirs and library_dirs
        # were set
        assert lib in cmd.library_dirs
        assert lib in cmd.rpath
        assert incl in cmd.include_dirs

    def test_optional_extension(self):
        # this extension will fail, but let's ignore this failure
        # with the optional argument.
        modules = [Extension('foo', ['xxx'], optional=False)]
        dist = Distribution({'name': 'xx', 'ext_modules': modules})
        cmd = self.build_ext(dist)
        cmd.ensure_finalized()
        with pytest.raises((UnknownFileError, CompileError)):
            cmd.run()  # should raise an error

        modules = [Extension('foo', ['xxx'], optional=True)]
        dist = Distribution({'name': 'xx', 'ext_modules': modules})
        cmd = self.build_ext(dist)
        cmd.ensure_finalized()
        cmd.run()  # should pass

    def test_finalize_options(self):
        # Make sure Python's include directories (for Python.h, pyconfig.h,
        # etc.) are in the include search path.
        modules = [Extension('foo', ['xxx'], optional=False)]
        dist = Distribution({'name': 'xx', 'ext_modules': modules})
        cmd = self.build_ext(dist)
        cmd.finalize_options()

        py_include = sysconfig.get_python_inc()
        for p in py_include.split(os.path.pathsep):
            assert p in cmd.include_dirs

        plat_py_include = sysconfig.get_python_inc(plat_specific=True)
        for p in plat_py_include.split(os.path.pathsep):
            assert p in cmd.include_dirs

        # make sure cmd.libraries is turned into a list
        # if it's a string
        cmd = self.build_ext(dist)
        cmd.libraries = 'my_lib, other_lib lastlib'
        cmd.finalize_options()
        assert cmd.libraries == ['my_lib', 'other_lib', 'lastlib']

        # make sure cmd.library_dirs is turned into a list
        # if it's a string
        cmd = self.build_ext(dist)
        cmd.library_dirs = f'my_lib_dir{os.pathsep}other_lib_dir'
        cmd.finalize_options()
        assert 'my_lib_dir' in cmd.library_dirs
        assert 'other_lib_dir' in cmd.library_dirs

        # make sure rpath is turned into a list
        # if it's a string
        cmd = self.build_ext(dist)
        cmd.rpath = f'one{os.pathsep}two'
        cmd.finalize_options()
        assert cmd.rpath == ['one', 'two']

        # make sure cmd.link_objects is turned into a list
        # if it's a string
        cmd = build_ext(dist)
        cmd.link_objects = 'one two,three'
        cmd.finalize_options()
        assert cmd.link_objects == ['one', 'two', 'three']

        # XXX more tests to perform for win32

        # make sure define is turned into 2-tuples
        # strings if they are ','-separated strings
        cmd = self.build_ext(dist)
        cmd.define = 'one,two'
        cmd.finalize_options()
        assert cmd.define == [('one', '1'), ('two', '1')]

        # make sure undef is turned into a list of
        # strings if they are ','-separated strings
        cmd = self.build_ext(dist)
        cmd.undef = 'one,two'
        cmd.finalize_options()
        assert cmd.undef == ['one', 'two']

        # make sure swig_opts is turned into a list
        cmd = self.build_ext(dist)
        cmd.swig_opts = None
        cmd.finalize_options()
        assert cmd.swig_opts == []

        cmd = self.build_ext(dist)
        cmd.swig_opts = '1 2'
        cmd.finalize_options()
        assert cmd.swig_opts == ['1', '2']

    def test_check_extensions_list(self):
        dist = Distribution()
        cmd = self.build_ext(dist)
        cmd.finalize_options()

        # 'extensions' option must be a list of Extension instances
        with pytest.raises(DistutilsSetupError):
            cmd.check_extensions_list('foo')

        # each element of 'ext_modules' option must be an
        # Extension instance or 2-tuple
        exts = [('bar', 'foo', 'bar'), 'foo']
        with pytest.raises(DistutilsSetupError):
            cmd.check_extensions_list(exts)

        # first element of each tuple in 'ext_modules'
        # must be the extension name (a string) and match
        # a python dotted-separated name
        exts = [('foo-bar', '')]
        with pytest.raises(DistutilsSetupError):
            cmd.check_extensions_list(exts)

        # second element of each tuple in 'ext_modules'
        # must be a dictionary (build info)
        exts = [('foo.bar', '')]
        with pytest.raises(DistutilsSetupError):
            cmd.check_extensions_list(exts)

        # ok this one should pass
        exts = [('foo.bar', {'sources': [''], 'libraries': 'foo', 'some': 'bar'})]
        cmd.check_extensions_list(exts)
        ext = exts[0]
        assert isinstance(ext, Extension)

        # check_extensions_list adds in ext the values passed
        # when they are in ('include_dirs', 'library_dirs', 'libraries'
        # 'extra_objects', 'extra_compile_args', 'extra_link_args')
        assert ext.libraries == 'foo'
        assert not hasattr(ext, 'some')

        # 'macros' element of build info dict must be 1- or 2-tuple
        exts = [
            (
                'foo.bar',
                {
                    'sources': [''],
                    'libraries': 'foo',
                    'some': 'bar',
                    'macros': [('1', '2', '3'), 'foo'],
                },
            )
        ]
        with pytest.raises(DistutilsSetupError):
            cmd.check_extensions_list(exts)

        exts[0][1]['macros'] = [('1', '2'), ('3',)]
        cmd.check_extensions_list(exts)
        assert exts[0].undef_macros == ['3']
        assert exts[0].define_macros == [('1', '2')]

    def test_get_source_files(self):
        modules = [Extension('foo', ['xxx'], optional=False)]
        dist = Distribution({'name': 'xx', 'ext_modules': modules})
        cmd = self.build_ext(dist)
        cmd.ensure_finalized()
        assert cmd.get_source_files() == ['xxx']

    def test_unicode_module_names(self):
        modules = [
            Extension('foo', ['aaa'], optional=False),
            Extension('föö', ['uuu'], optional=False),
        ]
        dist = Distribution({'name': 'xx', 'ext_modules': modules})
        cmd = self.build_ext(dist)
        cmd.ensure_finalized()
        assert re.search(r'foo(_d)?\..*', cmd.get_ext_filename(modules[0].name))
        assert re.search(r'föö(_d)?\..*', cmd.get_ext_filename(modules[1].name))
        assert cmd.get_export_symbols(modules[0]) == ['PyInit_foo']
        assert cmd.get_export_symbols(modules[1]) == ['PyInitU_f_1gaa']

    def test_compiler_option(self):
        # cmd.compiler is an option and
        # should not be overridden by a compiler instance
        # when the command is run
        dist = Distribution()
        cmd = self.build_ext(dist)
        cmd.compiler = 'unix'
        cmd.ensure_finalized()
        cmd.run()
        assert cmd.compiler == 'unix'

    def test_get_outputs(self):
        missing_compiler_executable()
        tmp_dir = self.mkdtemp()
        c_file = os.path.join(tmp_dir, 'foo.c')
        self.write_file(c_file, 'void PyInit_foo(void) {}\n')
        ext = Extension('foo', [c_file], optional=False)
        dist = Distribution({'name': 'xx', 'ext_modules': [ext]})
        cmd = self.build_ext(dist)
        fixup_build_ext(cmd)
        cmd.ensure_finalized()
        assert len(cmd.get_outputs()) == 1

        cmd.build_lib = os.path.join(self.tmp_dir, 'build')
        cmd.build_temp = os.path.join(self.tmp_dir, 'tempt')

        # issue #5977 : distutils build_ext.get_outputs
        # returns wrong result with --inplace
        other_tmp_dir = os.path.realpath(self.mkdtemp())
        old_wd = os.getcwd()
        os.chdir(other_tmp_dir)
        try:
            cmd.inplace = True
            cmd.run()
            so_file = cmd.get_outputs()[0]
        finally:
            os.chdir(old_wd)
        assert os.path.exists(so_file)
        ext_suffix = sysconfig.get_config_var('EXT_SUFFIX')
        assert so_file.endswith(ext_suffix)
        so_dir = os.path.dirname(so_file)
        assert so_dir == other_tmp_dir

        cmd.inplace = False
        cmd.compiler = None
        cmd.run()
        so_file = cmd.get_outputs()[0]
        assert os.path.exists(so_file)
        assert so_file.endswith(ext_suffix)
        so_dir = os.path.dirname(so_file)
        assert so_dir == cmd.build_lib

        # inplace = False, cmd.package = 'bar'
        build_py = cmd.get_finalized_command('build_py')
        build_py.package_dir = {'': 'bar'}
        path = cmd.get_ext_fullpath('foo')
        # checking that the last directory is the build_dir
        path = os.path.split(path)[0]
        assert path == cmd.build_lib

        # inplace = True, cmd.package = 'bar'
        cmd.inplace = True
        other_tmp_dir = os.path.realpath(self.mkdtemp())
        old_wd = os.getcwd()
        os.chdir(other_tmp_dir)
        try:
            path = cmd.get_ext_fullpath('foo')
        finally:
            os.chdir(old_wd)
        # checking that the last directory is bar
        path = os.path.split(path)[0]
        lastdir = os.path.split(path)[-1]
        assert lastdir == 'bar'

    def test_ext_fullpath(self):
        ext = sysconfig.get_config_var('EXT_SUFFIX')
        # building lxml.etree inplace
        # etree_c = os.path.join(self.tmp_dir, 'lxml.etree.c')
        # etree_ext = Extension('lxml.etree', [etree_c])
        # dist = Distribution({'name': 'lxml', 'ext_modules': [etree_ext]})
        dist = Distribution()
        cmd = self.build_ext(dist)
        cmd.inplace = True
        cmd.distribution.package_dir = {'': 'src'}
        cmd.distribution.packages = ['lxml', 'lxml.html']
        curdir = os.getcwd()
        wanted = os.path.join(curdir, 'src', 'lxml', 'etree' + ext)
        path = cmd.get_ext_fullpath('lxml.etree')
        assert wanted == path

        # building lxml.etree not inplace
        cmd.inplace = False
        cmd.build_lib = os.path.join(curdir, 'tmpdir')
        wanted = os.path.join(curdir, 'tmpdir', 'lxml', 'etree' + ext)
        path = cmd.get_ext_fullpath('lxml.etree')
        assert wanted == path

        # building twisted.runner.portmap not inplace
        build_py = cmd.get_finalized_command('build_py')
        build_py.package_dir = {}
        cmd.distribution.packages = ['twisted', 'twisted.runner.portmap']
        path = cmd.get_ext_fullpath('twisted.runner.portmap')
        wanted = os.path.join(curdir, 'tmpdir', 'twisted', 'runner', 'portmap' + ext)
        assert wanted == path

        # building twisted.runner.portmap inplace
        cmd.inplace = True
        path = cmd.get_ext_fullpath('twisted.runner.portmap')
        wanted = os.path.join(curdir, 'twisted', 'runner', 'portmap' + ext)
        assert wanted == path

    @pytest.mark.skipif('platform.system() != "Darwin"')
    @pytest.mark.usefixtures('save_env')
    def test_deployment_target_default(self):
        # Issue 9516: Test that, in the absence of the environment variable,
        # an extension module is compiled with the same deployment target as
        #  the interpreter.
        self._try_compile_deployment_target('==', None)

    @pytest.mark.skipif('platform.system() != "Darwin"')
    @pytest.mark.usefixtures('save_env')
    def test_deployment_target_too_low(self):
        # Issue 9516: Test that an extension module is not allowed to be
        # compiled with a deployment target less than that of the interpreter.
        with pytest.raises(DistutilsPlatformError):
            self._try_compile_deployment_target('>', '10.1')

    @pytest.mark.skipif('platform.system() != "Darwin"')
    @pytest.mark.usefixtures('save_env')
    def test_deployment_target_higher_ok(self):  # pragma: no cover
        # Issue 9516: Test that an extension module can be compiled with a
        # deployment target higher than that of the interpreter: the ext
        # module may depend on some newer OS feature.
        deptarget = sysconfig.get_config_var('MACOSX_DEPLOYMENT_TARGET')
        if deptarget:
            # increment the minor version number (i.e. 10.6 -> 10.7)
            deptarget = [int(x) for x in deptarget.split('.')]
            deptarget[-1] += 1
            deptarget = '.'.join(str(i) for i in deptarget)
            self._try_compile_deployment_target('<', deptarget)

    def _try_compile_deployment_target(self, operator, target):  # pragma: no cover
        if target is None:
            if os.environ.get('MACOSX_DEPLOYMENT_TARGET'):
                del os.environ['MACOSX_DEPLOYMENT_TARGET']
        else:
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = target

        jaraco.path.build(
            {
                'deptargetmodule.c': textwrap.dedent(f"""\
                    #include <AvailabilityMacros.h>

                    int dummy;

                    #if TARGET {operator} MAC_OS_X_VERSION_MIN_REQUIRED
                    #else
                    #error "Unexpected target"
                    #endif

                    """),
            },
            self.tmp_path,
        )

        # get the deployment target that the interpreter was built with
        target = sysconfig.get_config_var('MACOSX_DEPLOYMENT_TARGET')
        target = tuple(map(int, target.split('.')[0:2]))
        # format the target value as defined in the Apple
        # Availability Macros.  We can't use the macro names since
        # at least one value we test with will not exist yet.
        if target[:2] < (10, 10):
            # for 10.1 through 10.9.x -> "10n0"
            target = '%02d%01d0' % target
        else:
            # for 10.10 and beyond -> "10nn00"
            if len(target) >= 2:
                target = '%02d%02d00' % target
            else:
                # 11 and later can have no minor version (11 instead of 11.0)
                target = '%02d0000' % target
        deptarget_ext = Extension(
            'deptarget',
            [self.tmp_path / 'deptargetmodule.c'],
            extra_compile_args=[f'-DTARGET={target}'],
        )
        dist = Distribution({'name': 'deptarget', 'ext_modules': [deptarget_ext]})
        dist.package_dir = self.tmp_dir
        cmd = self.build_ext(dist)
        cmd.build_lib = self.tmp_dir
        cmd.build_temp = self.tmp_dir

        try:
            old_stdout = sys.stdout
            if not support.verbose:
                # silence compiler output
                sys.stdout = StringIO()
            try:
                cmd.ensure_finalized()
                cmd.run()
            finally:
                sys.stdout = old_stdout

        except CompileError:
            self.fail("Wrong deployment target during compilation")


class TestParallelBuildExt(TestBuildExt):
    def build_ext(self, *args, **kwargs):
        build_ext = super().build_ext(*args, **kwargs)
        build_ext.parallel = True
        return build_ext



from typing import Optional

import printerconstants


def center(
    text,                   # type: str
    cols=printerconstants.COLS     # type: Optional[int]
):
    # type: (...) -> str

    if not text:
        return text
    text = cut(text)
    miss = cols - len(text)
    if miss == 0:
        return text
    left = miss / 2
    return (" " * left) + text


def center_text(
    text,                       # type: str
    sep=" ",                    # type: str
    line_size=printerconstants.COLS    # type: Optional[int]
):
    # type: (...) -> str

    text = " %s " % text
    begin = sep * int((line_size - len(text)) / 2)
    middle = begin + text
    end = sep * (line_size - len(middle))

    return middle + end


def cut(
    s   # type: str
):
    # type: (...) -> str

    if len(s) > printerconstants.COLS:
        # If there is a line-feed at the end, keep it
        return s[:printerconstants.COLS] if (s[-1] != "\n") else (s[:(printerconstants.COLS - 1)] + "\n")
    return s


def get_text_ellipsis(
    text_id,        # type: str
    text,           # type: str
    padding_space   # type: int
):
    # type: (...) -> str

    remaining_space = printerconstants.COLS - (len(text_id) + padding_space)
    return "{}...".format(text[:remaining_space - 3]) if len(text) > remaining_space else "{}".format(text)



# -*- coding: utf-8 -*-
import json

import sysactions

from collections import OrderedDict

from api.order import get_order_service
from bustoken import TK_FISCALWRAPPER_CANCEL_ORDER
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from mw_helper import show_filtered_list_box_dialog
from posot import OrderTaker
from utilconfigurations import get_cancel_json
from utilfunctions import build_clean_customer_props_dict
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union
)

from .. import mb_context
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user


def void_order(
    pos_id,                             # type: Any[Union[int, str]]
    order_id='',                        # type: Optional[str]
    abandon=0,                          # type: Optional[int]
    void_reason=None,                   # type: Optional[Any[int, Dict]]
    pos_ot=None,                        # type: Optional[OrderTaker]
    new_order_status=None,              # type: Optional[int]
    void_type="VOID_CURRENT_ORDER",     # type: Optional[str]
):
    # type: (...) -> bool

    model = sysactions.get_model(pos_id)
    if pos_ot is None:
        pos_ot = sysactions.get_posot(model)

    order_id = int(order_id) if order_id != "" else ""
    void_order_id = order_id
    if void_reason is None:
        void_reason = get_void_reason_id(
            pos_id=pos_id,
            selected_reason="UNDEFINED_REASON",
            authentication_user="SYSTEM",
            void_type=void_type,
        )
    elif not isinstance(void_reason, dict):
        void_reason = get_fixed_void_reason(
            pos_id=pos_id,
            reason=void_reason,
            canceled_by="SYSTEM",
            void_type=void_type,
        )

    order_service = get_order_service(pos_id=pos_id)
    order_service.void_order(
        order_id=void_order_id,
        abandon=abandon != 0,
        order_status=new_order_status,
        void_reason=json.dumps(void_reason),
        blk_notify=pos_ot.blkopnotify,
    )

    clean_customer_dict = build_clean_customer_props_dict()

    pos_ot.setOrderCustomProperties(
        properties=clean_customer_dict,
        orderid=order_id,
    )

    return True


def void_fiscal_order(
    pos_id,                     # type: int
    order_id,                   # type: int
    justification,              # type: str
):
    # type: (...) -> None

    data = "|".join([str(pos_id), str(order_id), str(justification)])
    msg = mb_context.MB_EasySendMessage("FiscalWrapper", TK_FISCALWRAPPER_CANCEL_ORDER, format=FM_PARAM, data=data)
    if msg.token != TK_SYS_ACK:
        sysactions.show_messagebox(
            posid=pos_id,
            message="$ERROR_CANCEL_FISCAL_ORDER|{}".format(msg.data),
        )
        raise sysactions.StopAction(msg.data)


def get_void_reason_id(
    pos_id,                     # type: str
    selected_reason=None,       # type: Optional[str]
    void_type=None,             # type: Optional[str]
    authentication_user=None,   # type: Optional[str]
    is_void_line=False,         # type: Optional[bool]
):
    # type: (...) -> Optional[Dict]

    if not selected_reason:
        void_reason_options = _get_void_reason_to_show(pos_id=pos_id, void_type=void_type)

        total_void_reasons = len(void_reason_options)
        if total_void_reasons > 1:
            title = "$SELECT_REASON"
            selected_reason = show_filtered_list_box_dialog(
                pos_id=pos_id,
                options=void_reason_options,
                title=title,
                mask="NOFILTER",
            )
        elif total_void_reasons == 1:
            selected_reason = 0
            do_confirm_void(
                pos_id=pos_id,
                message="$CONFIRM_VOID_LINE" if is_void_line else "$CONFIRM_VOID_ORDER",
            )

        if selected_reason is None:
            return None

        selected_reason = int(selected_reason)
        void_reason_id = void_reason_options[selected_reason].split('-')[0].strip()
        void_reason_description = void_reason_options[selected_reason].split('-')[1].strip()
        void_reason_dict = {
            "VOID_REASON_ID": void_reason_id,
            "VOID_REASON_DESCR": void_reason_description,
            "VOID_TYPE": void_type,
        }
        if authentication_user not in [None, "None"]:
            void_reason_dict.update({"VOID_AUTH_USER": authentication_user})
    else:
        void_reason_dict = get_fixed_void_reason(
            pos_id=pos_id,
            reason=selected_reason,
            canceled_by=authentication_user,
            void_type=void_type,
        )

    return void_reason_dict


def do_confirm_void(
    pos_id,                         # type: str
    message="$CONFIRM_VOID_ORDER",  # type: str
    ignore_timeout=False,           # type: bool
):
    # type: (...) -> None

    confirmation = _get_confirmation_void_order(
        pos_id=pos_id,
        message=message,
    )
    accept_confirmations = [1] if ignore_timeout else [None, 1]
    if confirmation in accept_confirmations:
        raise sysactions.StopAction()


def get_fixed_void_reason(
    pos_id,         # type: str
    reason,         # type: str
    canceled_by,    # type: str
    void_type,      # type: str
):
    # type: (...) -> Dict

    void_reason = get_void_reason_dict().get(reason)
    void_reason_id = void_reason.get("id")
    void_reason_description = void_reason.get("reasonKey")

    model = sysactions.get_model(posid=pos_id)
    void_reason_description = sysactions.translate_message(
        model=model,
        message_key=void_reason_description.replace("$", ""),
    )

    void_reason_dict = {
        "VOID_REASON_ID": void_reason_id,
        "VOID_REASON_DESCR": void_reason_description,
        "VOID_TYPE": void_type,
    }
    if canceled_by not in [None, "None"]:
        void_reason_dict.update({"VOID_AUTH_USER": canceled_by})

    return void_reason_dict


def void_reason_list_to_ordered_dict(void_reasons):
    # type: (...) -> OrderedDict

    void_reasons = sorted(
        void_reasons,
        key=lambda k: k["id"]
    )
    ordered_dict = OrderedDict()
    for item in void_reasons:
        ordered_dict[item["reasonKey"].replace("$", "")] = item

    return ordered_dict


def get_void_reason_dict(
    filtered_items=None,  # type: Optional[List]
):
    # type: (...) -> Dict

    void_reason_list = [
        {"id": 101, "reasonKey": "$CHANGED_YOUR_MIND"},
        {"id": 102, "reasonKey": "$DUPLICATE"},
        {"id": 103, "reasonKey": "$WRONG_SALE"},
        {"id": 104, "reasonKey": "$CANCELLATION"},
        {"id": 105, "reasonKey": "$USER_CANCELLATION"},
        {"id": 106, "reasonKey": "$USER_FISCAL_CANCELLATION"},
        {"id": 107, "reasonKey": "$TIMEOUT_CANCELLATION"},
        {"id": 108, "reasonKey": "$UNDEFINED_REASON"},
        {"id": 109, "reasonKey": "$RECALL_ORDER_TO_OLD"},
        {"id": 110, "reasonKey": "$UNKNOWN_FISCAL_ERROR"},
        {"id": 111, "reasonKey": "$POS_MOBILE_ON_PAUSE"},
        {"id": 112, "reasonKey": "$MANAGER_CLEAR_STORED_ORDERS"},
    ]

    void_reason_list = void_reason_list_to_ordered_dict(void_reason_list)

    return _filter_void_reason_items(void_reason_list, filtered_items)


def _filter_void_reason_items(
    void_reasons,           # type: OrderedDict
    filtered_items=None,    # type: Optional[List]
):
    # type: (...) -> Dict

    if filtered_items:
        return {k: v for k, v in void_reasons.iteritems() if v.get("id") in filtered_items}

    return void_reasons


def _get_void_reason_to_show(
    pos_id,     # type: str
    void_type,  # type: str
):
    # type: (...) -> List[str]

    void_reason_dict = get_void_reason_dict(filtered_items=[101, 102, 103, 104])
    cancel_json = get_cancel_json()
    void_reason_key = {
        "VOID_CURRENT_ORDER": "orderReasons",
        "VOID_STORED_ORDER": "storedOrderReasons",
        "VOID_PAID_ORDER": "couponReasons",
        "VOID_LINE": "lineReasons",
        "VOID_LINE_ON_STORED_ORDER": "lineStoredOrderReasons"
    }
    if cancel_json and void_reason_key[void_type] in cancel_json:
        void_reason_dict = void_reason_list_to_ordered_dict(cancel_json[void_reason_key[void_type]])

    return _translate_array(pos_id, void_reason_dict)


def _translate_array(
    pos_id,    # type: str
    messages,  # type: dict
):
    # type: (...) -> List[str]

    model = sysactions.get_model(pos_id)
    result = []
    for key, item in messages.items():
        reason_id = item["id"]
        reason = item["reasonKey"].encode('utf-8')
        if reason.startswith("$"):
            reason = sysactions.translate_message(
                model=model,
                message_key=reason.replace("$", ""),
            )
        result.append("{} - {}".format(
            reason_id,
            reason,
        ))

    return result


def get_canceled_by(
    pos_id,       # type: str
    model,        # type: eTree.XML
    auth_user,    # type: Union[str]
):
    # type: (...) -> str

    authenticated_manager = get_menu_manager_authenticated_user(int(pos_id))
    if authenticated_manager is not None:
        return str(authenticated_manager)

    if auth_user:
        return str(auth_user)

    operator = model.find("Operator")
    return operator.get("id", "UNDEFINED") if operator is not None else "UNDEFINED"


def _get_confirmation_void_order(
    pos_id,     # type: str
    message,    # type: str
):
    # type: (...) -> Optional[int]

    return sysactions.show_messagebox(
        posid=pos_id,
        message=message,
        buttons="$OK|$CANCEL",
    )



import pyscripts

from mbcontextmessagehandler import MbContextMessageBus
from nreports_api import NReports


def get_nreports_service():
    # type: (...) -> NReports

    message_bus = MbContextMessageBus(mbcontext=pyscripts.mbcontext)
    return NReports(message_bus=message_bus)



from ._EftConstants import EftConstants
from util_eft import (
    get_tender_details,
    check_eft_transaction,
    get_electronic_payment_component_name,
    get_electronic_tenders,
    get_tender_id,
    has_electronic_payment,
    save_cancel_receipt,
    check_eft_is_valid,
    get_all_order_tenders,
)
from finish_eft_transactions import finish_eft_transactions
from request_eft import request_eft
from cancel_eft_transaction import (
    cancel_eft_transaction,
    clear_all_order_tenders,
)
from confirm_finish_eft_transactions import confirm_finish_eft_transactions
from cache_eft import (
    reprint_receipt,
    reprint_last_eft_receipt,
    set_eft_cache,
)
from abort_current_payment import abort_current_payment
from release_wait_confirmation import release_wait_confirmation



from typing import List

from ._LineParser import (
    LineParser,
    JSON,
)
from ._CommandLineParser import CommandLineParser


class Line(object):

    @staticmethod
    def parse(
        line,       # type: JSON
        columns,    # type: int
    ):
        # type: (...) -> List[str]

        if "command" in line:
            return CommandLineParser(columns=columns).parse(line=line)
        elif "text" in line:
            return LineParser(columns=columns).parse(line=line)

        return CommandLineParser(columns=columns).parse(
            line={
                "command": "{CMD_BREAK_LINE}"
            }
        )



from typing import (
    Any,
    Dict,
)


class LockBenefitRequest(object):
    def __init__(
        self,           # type: LockBenefitRequest
        benefit,        # type: Dict[str, Any]
        pos_id,         # type: int
        payment_info,   # type: Dict[str, Any]
    ):
        # type: (...) -> None

        self.benefit = benefit
        self.pos_id = pos_id
        self.payment_info = payment_info



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/resource.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19google/api/resource.proto\x12\ngoogle.api\x1a google/protobuf/descriptor.proto"\xee\x02\n\x12ResourceDescriptor\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x03(\t\x12\x12\n\nname_field\x18\x03 \x01(\t\x12\x37\n\x07history\x18\x04 \x01(\x0e\x32&.google.api.ResourceDescriptor.History\x12\x0e\n\x06plural\x18\x05 \x01(\t\x12\x10\n\x08singular\x18\x06 \x01(\t\x12\x33\n\x05style\x18\n \x03(\x0e\x32$.google.api.ResourceDescriptor.Style"[\n\x07History\x12\x17\n\x13HISTORY_UNSPECIFIED\x10\x00\x12\x1d\n\x19ORIGINALLY_SINGLE_PATTERN\x10\x01\x12\x18\n\x14\x46UTURE_MULTI_PATTERN\x10\x02"8\n\x05Style\x12\x15\n\x11STYLE_UNSPECIFIED\x10\x00\x12\x18\n\x14\x44\x45\x43LARATIVE_FRIENDLY\x10\x01"5\n\x11ResourceReference\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x12\n\nchild_type\x18\x02 \x01(\t:Y\n\x12resource_reference\x12\x1d.google.protobuf.FieldOptions\x18\x9f\x08 \x01(\x0b\x32\x1d.google.api.ResourceReference:Z\n\x13resource_definition\x12\x1c.google.protobuf.FileOptions\x18\x9d\x08 \x03(\x0b\x32\x1e.google.api.ResourceDescriptor:R\n\x08resource\x12\x1f.google.protobuf.MessageOptions\x18\x9d\x08 \x01(\x0b\x32\x1e.google.api.ResourceDescriptorBn\n\x0e\x63om.google.apiB\rResourceProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xf8\x01\x01\xa2\x02\x04GAPIb\x06proto3'
)


RESOURCE_REFERENCE_FIELD_NUMBER = 1055
resource_reference = DESCRIPTOR.extensions_by_name["resource_reference"]
RESOURCE_DEFINITION_FIELD_NUMBER = 1053
resource_definition = DESCRIPTOR.extensions_by_name["resource_definition"]
RESOURCE_FIELD_NUMBER = 1053
resource = DESCRIPTOR.extensions_by_name["resource"]

_RESOURCEDESCRIPTOR = DESCRIPTOR.message_types_by_name["ResourceDescriptor"]
_RESOURCEREFERENCE = DESCRIPTOR.message_types_by_name["ResourceReference"]
_RESOURCEDESCRIPTOR_HISTORY = _RESOURCEDESCRIPTOR.enum_types_by_name["History"]
_RESOURCEDESCRIPTOR_STYLE = _RESOURCEDESCRIPTOR.enum_types_by_name["Style"]
ResourceDescriptor = _reflection.GeneratedProtocolMessageType(
    "ResourceDescriptor",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESOURCEDESCRIPTOR,
        "__module__": "google.api.resource_pb2"
        # @@protoc_insertion_point(class_scope:google.api.ResourceDescriptor)
    },
)
_sym_db.RegisterMessage(ResourceDescriptor)

ResourceReference = _reflection.GeneratedProtocolMessageType(
    "ResourceReference",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESOURCEREFERENCE,
        "__module__": "google.api.resource_pb2"
        # @@protoc_insertion_point(class_scope:google.api.ResourceReference)
    },
)
_sym_db.RegisterMessage(ResourceReference)

if _descriptor._USE_C_DESCRIPTORS == False:
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(
        resource_reference
    )
    google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(
        resource_definition
    )
    google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(resource)

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016com.google.apiB\rResourceProtoP\001ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\370\001\001\242\002\004GAPI"
    _RESOURCEDESCRIPTOR._serialized_start = 76
    _RESOURCEDESCRIPTOR._serialized_end = 442
    _RESOURCEDESCRIPTOR_HISTORY._serialized_start = 293
    _RESOURCEDESCRIPTOR_HISTORY._serialized_end = 384
    _RESOURCEDESCRIPTOR_STYLE._serialized_start = 386
    _RESOURCEDESCRIPTOR_STYLE._serialized_end = 442
    _RESOURCEREFERENCE._serialized_start = 444
    _RESOURCEREFERENCE._serialized_end = 497
# @@protoc_insertion_point(module_scope)



from abc import ABCMeta, abstractmethod
from xml.etree import cElementTree

from ._Order import Order  # noqa
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)
from ._Table import Table  # noqa


class TableService(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def list_tables(self, pos_id):
        # type: (Union[int, str]) -> List[Table]
        raise NotImplementedError()

    @abstractmethod
    def start_service(self, pos_id, table_id, seats, tab_id="", mode="", source_table_id=""):
        # type: (Union[int, str], str, int, Optional[str], Optional[str], Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def recall_service(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def store_service(self, pos_id, table_id):
        # type: (Union[int, str], str) -> bool
        raise NotImplementedError()

    @abstractmethod
    def create_order(
        self,
        pos_id,                 # type: Union[int, str]
        table_id,               # type: str
        custom_properties,      # type: str
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_table_picture(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def get_selected_table(self, pos_id):
        # type: (Union[int, str]) -> Table
        raise NotImplementedError()

    @abstractmethod
    def get_table_info(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def do_total_table(self, pos_id, table_id, tip_rate=None, selected_to_total=None):
        # type: (Union[int, str], str, Optional[str], Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def reopen_table(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def close_table(
        self,           # type: TableService
        pos_id,         # type: str
        table_id,       # type: str
        user_id,        # type: str
    ):
        # type: (...) -> str

        raise NotImplementedError()

    @abstractmethod
    def set_table_available(self, pos_id, table_id):
        # type: (Union[int, str], str) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_current_table_id(self, pos_id):
        # type: (Union[int, str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def clear_service_tenders(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def change_service_tip(self, pos_id, table_id, percentage=None, amount=None):
        # type: (Union[int, str], str, Optional[str], Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def register_service_tender(
        self,                   # type: TableService
        pos_id,                 # type: str
        user_id,                # type: str
        table_id,               # type: str
        tender_id,              # type: int
        amount,                 # type: Any[float, str]
        tender_details="",      # type: str
    ):
        # type: (...) -> str

        raise NotImplementedError()

    @abstractmethod
    def get_table_orders(self, pos_id, table_id):
        # type: (Union[int, str], str) -> List[Order]
        raise NotImplementedError()

    @abstractmethod
    def set_parent_table(
        self,
        pos_id,             # type: str
        table_id,           # type: str
        parent_table_id,    # type: str
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def set_current_order(
        self,
        pos_id,     # type: Any[Union[str, int]]
        order_id,   # type: str
        table_id    # type: str
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_service_seats_count(self, table_info_xml):
        # type: (cElementTree.XML) -> int
        raise NotImplementedError()

    @abstractmethod
    def set_sale_line_seat(self, pos_id, seat_number, order_id, line_number, item_id, level, part_code):
        # type: (Union[int, str], int, str, str, str, str, str) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_sectors(
        self,
        pos_id,     # type: Any[Union[str, int]]
    ):
        # type: (...) -> List[str]

        raise NotImplementedError()

    @abstractmethod
    def get_operator_sector_info(
        self,
        pos_id,     # type: Any[Union[str, int]]
    ):
        # type: (...) -> List[str]

        raise NotImplementedError()

    @abstractmethod
    def update_operator_sector_info(
        self,
        pos_id,         # type: Any[Union[str, int]]
        update_info,    # type: Dict[str, Dict]
    ):
        # type: (...) -> List[str]

        raise NotImplementedError()

    @abstractmethod
    def set_table_custom_property(
        self,
        pos_id,  # type: Any[Union[str, int]]
        prop,           # type: str
        prop_value,     # type: str
        table_id,       # type: str
        service_id,     # type: int
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def set_table_custom_properties(
        self,
        pos_id,         # type: Any[Union[str, int]]
        properties,     # type: str
        table_id,       # type: str
        service_id,     # type: int
    ):
        # type: (...) -> None

        raise NotImplementedError()



from xml.etree.ElementTree import Element

from xml.etree import cElementTree as eTree

from messagebus import (
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)
from typing import (
    Optional,
    Dict,
)
from .token import (
    TK_POS_GET_MODEL,
    TK_POS_GETCONFIG,
)


def requires_model(f):
    def wrapper(*args, **kwargs):
        if args[0].model is None:
            args[0].model = args[0]._retrieve_new_model()

        return f(*args, **kwargs)

    return wrapper


class PosService(object):

    def __init__(
        self,               # type: PosService
        message_bus,        # type: MessageBus
        pos_id,             # type: int
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id

        self.component_name = "POS{}".format(self.pos_id)

        self.pos_config_cache = {}  # type: Dict[str, Optional[str]]

        self.model = None

    def get_model(self):
        # type: (...) -> Element

        if self.model is None:
            self.model = self._retrieve_new_model()

        return self.model

    def get_updated_model(self):
        # type: (...) -> Element

        self.model = self._retrieve_new_model()
        return self.model

    @requires_model
    def get_operator_session(self):
        # type: (...) -> Optional[str]

        for op in self.model.findall("Operator"):
            if op.get("state") == "LOGGEDIN":
                return op.get("sessionId")

        return None

    @requires_model
    def get_pod_type(self):
        # type: (...) -> str

        return self.model.find("WorkingMode").get("podType")

    @requires_model
    def get_pos_function(self):
        # type: (...) -> str

        return self.model.find("WorkingMode").get("posFunction")

    def get_hour_to_block_pos(
        self,    # type: PosService
    ):
        # type: (...) -> int

        hour_to_block_pos = self._get_pos_config_key(
            tag="[@name='hourToBlockPos']/integer",
        )
        return int(hour_to_block_pos or 0)

    def _get_pos_config_key(
        self,                 # type: PosService
        tag,                  # type: str
    ):
        # type: (...) -> Optional[str]

        if tag in self.pos_config_cache:
            return self.pos_config_cache[tag]

        message = Message(
            token=TK_POS_GETCONFIG,
            data_type=DataType.param,
        )
        response = self.message_bus.send_message(
            component=self.component_name,
            message=message,
        )
        if response.token != DefaultToken.TK_SYS_ACK:
            return None

        config = eTree.XML(response.data)
        complete_tag = "group[@name='POS']/group[@name='POS{}']/key{}".format(
            str(self.pos_id).zfill(2),
            tag,
        )
        tag_element = config.find(complete_tag)
        if tag_element is not None:
            self.pos_config_cache[tag] = tag_element.text
            return self.pos_config_cache[tag]

        self.pos_config_cache[tag] = None
        return self.pos_config_cache[tag]

    def _retrieve_new_model(self):
        # type: (...) -> Element

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_POS_GET_MODEL,
                data_type=DataType.string,
                data=str(self.pos_id).encode()
            )
        )
        if msg.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Could not retrieve state of pos '{}'".format(self.pos_id))

        return eTree.XML(msg.data)



from typing import Optional, Union, List, Dict

from ._Comment import Comment
from ._CustomStorage import CustomStorage
from ._Item import Item
from ._ProdStates import ProdStates
from ._ProductionOrder import ProductionOrder
from ._StateEvent import StateEvent
from ._TagEvent import TagEvent
from ._TagEventType import TagEventType
from ._User import User
from ._Course import Course
from ._ProductionModelJsonEncoder import ProductionModelJsonEncoder
from ._ProductionExceptions import (
    TagNotFoundException,
    EventNotFoundException,
    PrinterNameFoundException,
    ViewNameFoundException,
    CommandNotFoundException,
    BuilderNotFoundException,
    ReprocessTimeNotFoundException,
    NoBoxesConfiguredException,
    NoViewsConfiguredException,
    ReprocessInvalidTimeException,
    PickListOrderNotFoundException,
    InvalidOrderToPrint,
    JitNotFoundException,
)
from ._PicklistCleanerManagerConfig import PicklistCleanerManagerConfig


def find_order_item(order, item):
    # type: (ProductionOrder, Union[Item, str]) -> Optional[Item]
    if isinstance(item, Item):
        line_id = item.get_line_id()
    else:
        line_id = item
    for item in order.items:
        if item.get_line_id() == line_id:
            return item
    return None


def get_current_line_id(line):
    return str(line.line_number) + "-" + line.item_id + "-" + str(int(line.level)) + "-" + str(line.part_code)


def has_tag(tags, line_id, tag):
    # type: (Dict, str, str) -> bool
    return line_id in tags and tag in tags[line_id]


def add_tag(tags, line_id, tag):
    # type: (Dict[str, List[str]], str, str) -> None
    if line_id in tags:
        tags[line_id].append(tag)
    else:
        tags[line_id] = [tag]


def create_states_from_state(state):
    state_order = ["IN_PROGRESS", "STORED", "TOTALED", "PAID", "VOIDED", "ABANDONED"]

    states = []

    try:
        index = state_order.index(state)
        for state in state_order[index:]:
            states.append(state)
    except ValueError:
        pass

    return states


def is_product(item):
    # type: (Item) -> bool
    return item.item_type in ("PRODUCT", "CANADD")



import json

from application.interactor.end_of_sale_notes import EndOfSaleNotesInteractor
from application.interactor.end_of_sale_notes.dto import EndOfSaleNotesRequest
from application.model import AsyncUuidMessageProcessor
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import (
    DataType,
    DefaultToken,
    Event,
    Message,
)
from typing import (
    Dict,
    Optional,
)


class EndOfSaleNotesProcessor(AsyncUuidMessageProcessor):

    def __init__(
        self,           # type: EndOfSaleNotesProcessor
        interactor,     # type: EndOfSaleNotesInteractor
    ):
        # type: (...) -> None

        super(EndOfSaleNotesProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,           # type: EndOfSaleNotesProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> EndOfSaleNotesRequest

        request_json = json.loads(data)
        return EndOfSaleNotesRequest(
            pos_id=request_json["posId"],
            order_id=request_json["orderId"],
        )

    def call_business(
        self,       # type: EndOfSaleNotesProcessor
        request,    # type: EndOfSaleNotesRequest
    ):
        # type: (...) -> Optional[Dict[str, str]]

        return self.interactor.execute(request=request)

    def format_response(
        self,           # type: EndOfSaleNotesProcessor
        message_bus,    # type: MbContextMessageBus
        message,        # type: Message
        event,          # type: Event
        request,        # type: None
        response,       # type: str
    ):
        # type: (...) -> None

        if response:
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=DefaultToken.TK_SYS_ACK,
                    data_type=DataType.json,
                    data=response,
                ),
            )
        else:
            message_bus.reply_message(
                message=message,
                reply_message=Message(token=DefaultToken.TK_SYS_NAK),
            )

    def format_exception(
        self,           # type: EndOfSaleNotesProcessor
        message_bus,    # type: MbContextMessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: None
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=Message(
                token=DefaultToken.TK_SYS_NAK,
                data_type=DataType.json,
                data=str(exception),
            ),
        )

    def format_parse_exception(
        self,           # type: EndOfSaleNotesProcessor
        message_bus,    # type: MbContextMessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: None
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception,
        )



from mw_helper import ensure_iterable

from ._OrderCondition import OrderCondition
from production.model import ProductionOrder


class ProdStateCondition(OrderCondition):
    def __init__(self, prod_states):
        prod_states = ensure_iterable(prod_states)

        self.valid_prod_states = {}
        for prod_state in prod_states:
            self.valid_prod_states[prod_state] = prod_state

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        return order.prod_state in self.valid_prod_states



from requests import Response


class ErrorResponseUtil(object):

    @staticmethod
    def get_error_response(response, message_not_found="$NOT_FOUND"):
        # type: (Response, str) -> str

        if response is not None:
            if response.status_code in [400, 403, 404]:
                return message_not_found
            if response.status_code == 503:
                return "$CONNECTION_NOT_FOUND"
            if response.status_code == 412:
                return "$INVALID_BOH_CONFIGURATION"
            if response.reason is not None:
                return response.reason

        return "$UNKNOWN_ERROR"



import glob
import os
from typing import List

from filesystem import FileSystem


class RealFileSystem(FileSystem):

    def list_entries(
        self,
        path: str,
    ) -> List[str]:

        ret = []
        for entry in os.listdir(path):
            ret.append(os.path.join(path, entry))

        return ret

    def read_file_content(
        self,
        path: str,
    ) -> bytes:

        with open(path, "rb") as f:
            return f.read()

    def is_dir(
        self,
        path: str,
    ) -> bool:

        return os.path.isdir(path)

    def is_file(
        self,
        path: str,
    ) -> bool:

        return os.path.isfile(path)



from typing import List

from model.customexception import (
    MessageBusError,
    UpdateRuptureException,
)
import json

import sysactions
from msgbus import TK_SYS_ACK

from api.product import get_product_service
from bustoken import TK_SLCTRL_PMGR_RUPTURE_GET_DISABLED


class Rupture(object):

    @staticmethod
    def get_disabled_products():
        # type: (...) -> List[int]

        msg = sysactions.send_message("ORDERMGR", TK_SLCTRL_PMGR_RUPTURE_GET_DISABLED)

        if msg.token != TK_SYS_ACK:
            raise MessageBusError(
                component="ORDERMGR",
                token=msg.token,
                data=msg.data,
            )

        return json.loads(msg.data.decode("utf-8"))

    @staticmethod
    def update_items(
        user_id,            # type: int
        parts_to_enable,    # type: List[int]
        parts_to_disable,   # type: List[int]
    ):
        # type: (...) -> None

        product_service = get_product_service()
        try:
            product_service.update_items(
                user_id=user_id,
                enabled=parts_to_enable,
                disabled=parts_to_disable,
            )
        except (Exception,) as ex:
            raise UpdateRuptureException(ex)



from xml.etree import cElementTree as eTree
from typing import (
    Optional,
    List,
)

from messagebus import (
    DefaultToken,
    Message,
    MessageBus,
    TokenCreator,
    TokenPriority,
)

TK_STORECFG_GETFULL = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="13",
    number="2",
)  # 19922946


class StoreConfigRepository(object):

    def __init__(
        self,
        message_bus: MessageBus,
    ) -> None:

        self._message_bus = message_bus
        self._store_wide_config: Optional[eTree.Element] = None
        self._available_tips: Optional[List[str]] = []
        self._tip_rate: Optional[int] = None
        self._store_id: Optional[str] = None

    def get_store_wide_config(
        self,
    ) -> eTree.Element:

        if self._store_wide_config:
            return self._store_wide_config

        msg = self._message_bus.send_message(
            component="StoreWideConfig",
            message=Message(token=TK_STORECFG_GETFULL),
        )
        if msg.token == DefaultToken.TK_SYS_ACK:
            self._store_wide_config = eTree.XML(msg.data.decode())
            return self._store_wide_config

        raise RuntimeError("StoreWideConfig not found")

    def get_available_tips(
        self,
    ) -> List[str]:

        if self._available_tips:
            return self._available_tips

        store_wide_config = self.get_store_wide_config()
        tag_to_find = self._get_full_name_tag_from_store_tag(tag="key[@name='AvailableTips']/array/string")
        available_tips = store_wide_config.findall(tag_to_find)
        for tip in available_tips:
            self._available_tips.append(tip.text)

        return self._available_tips

    def get_store_id(
        self,
    ) -> str:

        if self._store_id:
            return self._store_id

        store_wide_config = self.get_store_wide_config()
        tag_to_find = self._get_full_name_tag_from_store_tag(tag="key[@name='Id']/string")
        self._store_id = store_wide_config.find(tag_to_find)

        return self._store_id.text

    def get_tip_rate(
        self,
    ) -> float:

        if self._tip_rate:
            return self._tip_rate

        store_wide_config = self.get_store_wide_config()
        tag_to_find = self._get_full_name_tag_from_store_tag(tag="key[@name='TipRate']/string")
        self._tip_rate = float(store_wide_config.find(tag_to_find).text or 10)
        return self._tip_rate

    @staticmethod
    def _get_full_name_tag_from_store_tag(
        tag: str,
    ) -> str:

        return f"group[@name='PublishedConfiguration']/group[@name='Store']/{tag}"



# iban.py - functions for handling Spanish IBANs
# coding: utf-8
#
# Copyright (C) 2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Spanish IBAN (International Bank Account Number).

The IBAN is used to identify bank accounts across national borders. The
Spanish IBAN is built up of the IBAN prefix (ES) and check digits, followed
by the 20 digit CCC (Código Cuenta Corriente).

>>> validate('ES77 1234-1234-16 1234567890')
'ES7712341234161234567890'
>>> to_ccc('ES77 1234-1234-16 1234567890')
'12341234161234567890'
>>> format('ES771234-1234-16 1234567890')
'ES77 1234 1234 1612 3456 7890'
>>> validate('GR1601101050000010547023795')  # different country
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('ES12 1234-1234-16 1234567890')  # invalid IBAN check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('ES15 1234-1234-17 1234567890')  # invalid CCC check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum import iban
from stdnum.es import ccc
from stdnum.exceptions import *


__all__ = ['compact', 'format', 'to_ccc', 'validate', 'is_valid']


compact = iban.compact
format = iban.format


def to_ccc(number):
    """Return the CCC (Código Cuenta Corriente) part of the number."""
    number = compact(number)
    if not number.startswith('ES'):
        raise InvalidComponent()
    return number[4:]


def validate(number):
    """Check if the number provided is a valid Spanish IBAN."""
    number = iban.validate(number, check_country=False)
    ccc.validate(to_ccc(number))
    return number


def is_valid(number):
    """Check if the number provided is a valid Spanish IBAN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import iso8601

from ._OrderChangerProductionBox import OrderChangerProductionBox
from mw_helper import convert_from_localtime_to_utc


class FilterDeletedItemsBeforeFirstStateBox(OrderChangerProductionBox):
    def __init__(self, name, sons, order_state, logger=None):
        super(FilterDeletedItemsBeforeFirstStateBox, self).__init__(name, sons, logger)
        self.order_state = order_state
        self.first_stored_time = None

    def change_order(self, order):
        self.first_stored_time = self._get_first_state_time(order)
        if self.first_stored_time is None:
            return order

        self._remove_empty_items(order.items)
        return order

    def _get_first_state_time(self, order):
        stored_states = [state for state in order.state_history if state.state == self.order_state]
        if len(stored_states) > 0:
            return convert_from_localtime_to_utc(iso8601.parse_date(stored_states[0].timestamp)).isoformat()

        return None

    def _remove_empty_items(self, items):
        for item in items[:]:
            if item.last_updated and item.last_updated > self.first_stored_time:
                continue

            if item.qty == 0:
                item_is_not_ingredient_or_option = item.item_type != 'INGREDIENT' and item.item_type != 'OPTION'
                if item_is_not_ingredient_or_option:
                    items.remove(item)
                else:
                    item.show_removed = True

            if item.items and len(item.items) > 0:
                self._remove_empty_items(item.items)



from helper import ExtendedEnum


class ProductTypeId(ExtendedEnum):
    REGULAR = 0
    OPTION = 1
    COMBO = 2
    MENU = 3
    SET = 4
    PROMO = 5
    COUPON = 6


class ProductTypeName:
    OPTION = "OPTION"
    COMBO = "COMBO"
    MENU = "MENU"
    SET = "SET"
    PROMO = "PROMO"
    COUPON = "COUPON"
    PRODUCT = "PRODUCT"
    CANADD = "CANADD"
    INGREDIENT = "INGREDIENT"



"""
Create a dist_info directory
As defined in the wheel specification
"""

import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import cast

from .. import _normalization
from .egg_info import egg_info as egg_info_cls

from distutils import log
from distutils.core import Command


class dist_info(Command):
    """
    This command is private and reserved for internal use of setuptools,
    users should rely on ``setuptools.build_meta`` APIs.
    """

    description = "DO NOT CALL DIRECTLY, INTERNAL ONLY: create .dist-info directory"

    user_options = [
        (
            'output-dir=',
            'o',
            "directory inside of which the .dist-info will be"
            "created [default: top of the source tree]",
        ),
        ('tag-date', 'd', "Add date stamp (e.g. 20050528) to version number"),
        ('tag-build=', 'b', "Specify explicit tag to add to version number"),
        ('no-date', 'D', "Don't include date stamp [default]"),
        ('keep-egg-info', None, "*TRANSITIONAL* will be removed in the future"),
    ]

    boolean_options = ['tag-date', 'keep-egg-info']
    negative_opt = {'no-date': 'tag-date'}

    def initialize_options(self):
        self.output_dir = None
        self.name = None
        self.dist_info_dir = None
        self.tag_date = None
        self.tag_build = None
        self.keep_egg_info = False

    def finalize_options(self):
        dist = self.distribution
        project_dir = dist.src_root or os.curdir
        self.output_dir = Path(self.output_dir or project_dir)

        egg_info = cast(egg_info_cls, self.reinitialize_command("egg_info"))
        egg_info.egg_base = str(self.output_dir)

        if self.tag_date:
            egg_info.tag_date = self.tag_date
        else:
            self.tag_date = egg_info.tag_date

        if self.tag_build:
            egg_info.tag_build = self.tag_build
        else:
            self.tag_build = egg_info.tag_build

        egg_info.finalize_options()
        self.egg_info = egg_info

        name = _normalization.safer_name(dist.get_name())
        version = _normalization.safer_best_effort_version(dist.get_version())
        self.name = f"{name}-{version}"
        self.dist_info_dir = os.path.join(self.output_dir, f"{self.name}.dist-info")

    @contextmanager
    def _maybe_bkp_dir(self, dir_path: str, requires_bkp: bool):
        if requires_bkp:
            bkp_name = f"{dir_path}.__bkp__"
            _rm(bkp_name, ignore_errors=True)
            shutil.copytree(dir_path, bkp_name, dirs_exist_ok=True, symlinks=True)
            try:
                yield
            finally:
                _rm(dir_path, ignore_errors=True)
                shutil.move(bkp_name, dir_path)
        else:
            yield

    def run(self):
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.egg_info.run()
        egg_info_dir = self.egg_info.egg_info
        assert os.path.isdir(egg_info_dir), ".egg-info dir should have been created"

        log.info("creating '{}'".format(os.path.abspath(self.dist_info_dir)))
        bdist_wheel = self.get_finalized_command('bdist_wheel')

        # TODO: if bdist_wheel if merged into setuptools, just add "keep_egg_info" there
        with self._maybe_bkp_dir(egg_info_dir, self.keep_egg_info):
            bdist_wheel.egg2dist(egg_info_dir, self.dist_info_dir)


def _rm(dir_name, **opts):
    if os.path.isdir(dir_name):
        shutil.rmtree(dir_name, **opts)



from .api import _render_migration_diffs
from .api import compare_metadata
from .api import produce_migrations
from .api import render_python_code
from .api import RevisionContext
from .compare import _produce_net_changes
from .compare import comparators
from .render import render_op_text
from .render import renderers
from .rewriter import Rewriter



#!/usr/bin/env python3

# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import subprocess
import sys

import pkg_resources

from opentelemetry.instrumentation.bootstrap_gen import (
    default_instrumentations,
    libraries,
)
from opentelemetry.instrumentation.version import __version__

logger = logging.getLogger(__name__)


def _syscall(func):
    def wrapper(package=None):
        try:
            if package:
                return func(package)
            return func()
        except subprocess.SubprocessError as exp:
            cmd = getattr(exp, "cmd", None)
            if cmd:
                msg = f'Error calling system command "{" ".join(cmd)}"'
            if package:
                msg = f'{msg} for package "{package}"'
            raise RuntimeError(msg)

    return wrapper


@_syscall
def _sys_pip_install(package):
    # explicit upgrade strategy to override potential pip config
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "-U",
            "--upgrade-strategy",
            "only-if-needed",
            package,
        ]
    )


def _pip_check():
    """Ensures none of the instrumentations have dependency conflicts.
    Clean check reported as:
    'No broken requirements found.'
    Dependency conflicts are reported as:
    'opentelemetry-instrumentation-flask 1.0.1 has requirement opentelemetry-sdk<2.0,>=1.0, but you have opentelemetry-sdk 0.5.'
    To not be too restrictive, we'll only check for relevant packages.
    """
    with subprocess.Popen(
        [sys.executable, "-m", "pip", "check"], stdout=subprocess.PIPE
    ) as check_pipe:
        pip_check = check_pipe.communicate()[0].decode()
        pip_check_lower = pip_check.lower()
    for package_tup in libraries.values():
        for package in package_tup:
            if package.lower() in pip_check_lower:
                raise RuntimeError(f"Dependency conflict found: {pip_check}")


def _is_installed(req):
    if req in sys.modules:
        return True

    try:
        pkg_resources.get_distribution(req)
    except pkg_resources.DistributionNotFound:
        return False
    except pkg_resources.VersionConflict as exc:
        logger.warning(
            "instrumentation for package %s is available but version %s is installed. Skipping.",
            exc.req,
            exc.dist.as_requirement(),  # pylint: disable=no-member
        )
        return False
    return True


def _find_installed_libraries():
    libs = default_instrumentations[:]
    libs.extend(
        [
            v["instrumentation"]
            for _, v in libraries.items()
            if _is_installed(v["library"])
        ]
    )
    return libs


def _run_requirements():
    logger.setLevel(logging.ERROR)
    print("\n".join(_find_installed_libraries()), end="")


def _run_install():
    for lib in _find_installed_libraries():
        _sys_pip_install(lib)
    _pip_check()


def run() -> None:
    action_install = "install"
    action_requirements = "requirements"

    parser = argparse.ArgumentParser(
        description="""
        opentelemetry-bootstrap detects installed libraries and automatically
        installs the relevant instrumentation packages for them.
        """
    )
    parser.add_argument(
        "--version",
        help="print version information",
        action="version",
        version="%(prog)s " + __version__,
    )
    parser.add_argument(
        "-a",
        "--action",
        choices=[action_install, action_requirements],
        default=action_requirements,
        help="""
        install - uses pip to install the new requirements using to the
                  currently active site-package.
        requirements - prints out the new requirements to stdout. Action can
                       be piped and appended to a requirements.txt file.
        """,
    )
    args = parser.parse_args()

    cmd = {
        action_install: _run_install,
        action_requirements: _run_requirements,
    }[args.action]
    cmd()



from helper import ExtendedEnum
from messagebus import TokenCreator, TokenPriority

MESSAGE_GROUP = "95"


class ListenedTokens(ExtendedEnum):
    TK_POS_UPDATER_PERFORM_CATALOG_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_POS_UPDATER_PERFORM_MEDIA_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")
    TK_POS_UPDATER_PERFORM_USER_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "3")
    TK_POS_GET_CATALOG_VERSION_TO_APPLY = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "4")
    TK_POS_UPDATER_PERFORM_LOADER_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "5")
    TK_POS_UPDATER_GET_VERSIONS = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "6")
    TK_POS_UPDATER_PERFORM_BINARY_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "7")
    TK_POS_HAS_BINARY_UPDATE_TO_APPLY = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "8")
    TK_POS_UPDATER_PERFORM_CHROME_UPDATE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "9")
    TK_POS_UPDATER_HAS_PENDING_LOADER_UPDATE = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=MESSAGE_GROUP,
        number="10",
    )



import os

import cfgtools

default_delivery_type = None


def get_default_delivery_type():
    # type: (...) -> str

    global default_delivery_type

    if default_delivery_type is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        default_delivery_type = config.find_value("Customizations.DefaultDeliveryType", "delivery")

    return default_delivery_type.lower()



data = (
'tyal',    # 0x00
'tyalg',    # 0x01
'tyalm',    # 0x02
'tyalb',    # 0x03
'tyals',    # 0x04
'tyalt',    # 0x05
'tyalp',    # 0x06
'tyalh',    # 0x07
'tyam',    # 0x08
'tyab',    # 0x09
'tyabs',    # 0x0a
'tyas',    # 0x0b
'tyass',    # 0x0c
'tyang',    # 0x0d
'tyaj',    # 0x0e
'tyac',    # 0x0f
'tyak',    # 0x10
'tyat',    # 0x11
'tyap',    # 0x12
'tyah',    # 0x13
'tyae',    # 0x14
'tyaeg',    # 0x15
'tyaegg',    # 0x16
'tyaegs',    # 0x17
'tyaen',    # 0x18
'tyaenj',    # 0x19
'tyaenh',    # 0x1a
'tyaed',    # 0x1b
'tyael',    # 0x1c
'tyaelg',    # 0x1d
'tyaelm',    # 0x1e
'tyaelb',    # 0x1f
'tyaels',    # 0x20
'tyaelt',    # 0x21
'tyaelp',    # 0x22
'tyaelh',    # 0x23
'tyaem',    # 0x24
'tyaeb',    # 0x25
'tyaebs',    # 0x26
'tyaes',    # 0x27
'tyaess',    # 0x28
'tyaeng',    # 0x29
'tyaej',    # 0x2a
'tyaec',    # 0x2b
'tyaek',    # 0x2c
'tyaet',    # 0x2d
'tyaep',    # 0x2e
'tyaeh',    # 0x2f
'teo',    # 0x30
'teog',    # 0x31
'teogg',    # 0x32
'teogs',    # 0x33
'teon',    # 0x34
'teonj',    # 0x35
'teonh',    # 0x36
'teod',    # 0x37
'teol',    # 0x38
'teolg',    # 0x39
'teolm',    # 0x3a
'teolb',    # 0x3b
'teols',    # 0x3c
'teolt',    # 0x3d
'teolp',    # 0x3e
'teolh',    # 0x3f
'teom',    # 0x40
'teob',    # 0x41
'teobs',    # 0x42
'teos',    # 0x43
'teoss',    # 0x44
'teong',    # 0x45
'teoj',    # 0x46
'teoc',    # 0x47
'teok',    # 0x48
'teot',    # 0x49
'teop',    # 0x4a
'teoh',    # 0x4b
'te',    # 0x4c
'teg',    # 0x4d
'tegg',    # 0x4e
'tegs',    # 0x4f
'ten',    # 0x50
'tenj',    # 0x51
'tenh',    # 0x52
'ted',    # 0x53
'tel',    # 0x54
'telg',    # 0x55
'telm',    # 0x56
'telb',    # 0x57
'tels',    # 0x58
'telt',    # 0x59
'telp',    # 0x5a
'telh',    # 0x5b
'tem',    # 0x5c
'teb',    # 0x5d
'tebs',    # 0x5e
'tes',    # 0x5f
'tess',    # 0x60
'teng',    # 0x61
'tej',    # 0x62
'tec',    # 0x63
'tek',    # 0x64
'tet',    # 0x65
'tep',    # 0x66
'teh',    # 0x67
'tyeo',    # 0x68
'tyeog',    # 0x69
'tyeogg',    # 0x6a
'tyeogs',    # 0x6b
'tyeon',    # 0x6c
'tyeonj',    # 0x6d
'tyeonh',    # 0x6e
'tyeod',    # 0x6f
'tyeol',    # 0x70
'tyeolg',    # 0x71
'tyeolm',    # 0x72
'tyeolb',    # 0x73
'tyeols',    # 0x74
'tyeolt',    # 0x75
'tyeolp',    # 0x76
'tyeolh',    # 0x77
'tyeom',    # 0x78
'tyeob',    # 0x79
'tyeobs',    # 0x7a
'tyeos',    # 0x7b
'tyeoss',    # 0x7c
'tyeong',    # 0x7d
'tyeoj',    # 0x7e
'tyeoc',    # 0x7f
'tyeok',    # 0x80
'tyeot',    # 0x81
'tyeop',    # 0x82
'tyeoh',    # 0x83
'tye',    # 0x84
'tyeg',    # 0x85
'tyegg',    # 0x86
'tyegs',    # 0x87
'tyen',    # 0x88
'tyenj',    # 0x89
'tyenh',    # 0x8a
'tyed',    # 0x8b
'tyel',    # 0x8c
'tyelg',    # 0x8d
'tyelm',    # 0x8e
'tyelb',    # 0x8f
'tyels',    # 0x90
'tyelt',    # 0x91
'tyelp',    # 0x92
'tyelh',    # 0x93
'tyem',    # 0x94
'tyeb',    # 0x95
'tyebs',    # 0x96
'tyes',    # 0x97
'tyess',    # 0x98
'tyeng',    # 0x99
'tyej',    # 0x9a
'tyec',    # 0x9b
'tyek',    # 0x9c
'tyet',    # 0x9d
'tyep',    # 0x9e
'tyeh',    # 0x9f
'to',    # 0xa0
'tog',    # 0xa1
'togg',    # 0xa2
'togs',    # 0xa3
'ton',    # 0xa4
'tonj',    # 0xa5
'tonh',    # 0xa6
'tod',    # 0xa7
'tol',    # 0xa8
'tolg',    # 0xa9
'tolm',    # 0xaa
'tolb',    # 0xab
'tols',    # 0xac
'tolt',    # 0xad
'tolp',    # 0xae
'tolh',    # 0xaf
'tom',    # 0xb0
'tob',    # 0xb1
'tobs',    # 0xb2
'tos',    # 0xb3
'toss',    # 0xb4
'tong',    # 0xb5
'toj',    # 0xb6
'toc',    # 0xb7
'tok',    # 0xb8
'tot',    # 0xb9
'top',    # 0xba
'toh',    # 0xbb
'twa',    # 0xbc
'twag',    # 0xbd
'twagg',    # 0xbe
'twags',    # 0xbf
'twan',    # 0xc0
'twanj',    # 0xc1
'twanh',    # 0xc2
'twad',    # 0xc3
'twal',    # 0xc4
'twalg',    # 0xc5
'twalm',    # 0xc6
'twalb',    # 0xc7
'twals',    # 0xc8
'twalt',    # 0xc9
'twalp',    # 0xca
'twalh',    # 0xcb
'twam',    # 0xcc
'twab',    # 0xcd
'twabs',    # 0xce
'twas',    # 0xcf
'twass',    # 0xd0
'twang',    # 0xd1
'twaj',    # 0xd2
'twac',    # 0xd3
'twak',    # 0xd4
'twat',    # 0xd5
'twap',    # 0xd6
'twah',    # 0xd7
'twae',    # 0xd8
'twaeg',    # 0xd9
'twaegg',    # 0xda
'twaegs',    # 0xdb
'twaen',    # 0xdc
'twaenj',    # 0xdd
'twaenh',    # 0xde
'twaed',    # 0xdf
'twael',    # 0xe0
'twaelg',    # 0xe1
'twaelm',    # 0xe2
'twaelb',    # 0xe3
'twaels',    # 0xe4
'twaelt',    # 0xe5
'twaelp',    # 0xe6
'twaelh',    # 0xe7
'twaem',    # 0xe8
'twaeb',    # 0xe9
'twaebs',    # 0xea
'twaes',    # 0xeb
'twaess',    # 0xec
'twaeng',    # 0xed
'twaej',    # 0xee
'twaec',    # 0xef
'twaek',    # 0xf0
'twaet',    # 0xf1
'twaep',    # 0xf2
'twaeh',    # 0xf3
'toe',    # 0xf4
'toeg',    # 0xf5
'toegg',    # 0xf6
'toegs',    # 0xf7
'toen',    # 0xf8
'toenj',    # 0xf9
'toenh',    # 0xfa
'toed',    # 0xfb
'toel',    # 0xfc
'toelg',    # 0xfd
'toelm',    # 0xfe
'toelb',    # 0xff
)



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/gapic/metadata/gapic_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*google/gapic/metadata/gapic_metadata.proto\x12\x15google.gapic.metadata"\xf0\x05\n\rGapicMetadata\x12\x0e\n\x06schema\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x10\n\x08language\x18\x03 \x01(\t\x12\x15\n\rproto_package\x18\x04 \x01(\t\x12\x17\n\x0flibrary_package\x18\x05 \x01(\t\x12\x44\n\x08services\x18\x06 \x03(\x0b\x32\x32.google.gapic.metadata.GapicMetadata.ServicesEntry\x1ai\n\rServicesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12G\n\x05value\x18\x02 \x01(\x0b\x32\x38.google.gapic.metadata.GapicMetadata.ServiceForTransport:\x02\x38\x01\x1a\xd3\x01\n\x13ServiceForTransport\x12V\n\x07\x63lients\x18\x01 \x03(\x0b\x32\x45.google.gapic.metadata.GapicMetadata.ServiceForTransport.ClientsEntry\x1a\x64\n\x0c\x43lientsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x43\n\x05value\x18\x02 \x01(\x0b\x32\x34.google.gapic.metadata.GapicMetadata.ServiceAsClient:\x02\x38\x01\x1a\xd5\x01\n\x0fServiceAsClient\x12\x16\n\x0elibrary_client\x18\x01 \x01(\t\x12L\n\x04rpcs\x18\x02 \x03(\x0b\x32>.google.gapic.metadata.GapicMetadata.ServiceAsClient.RpcsEntry\x1a\\\n\tRpcsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12>\n\x05value\x18\x02 \x01(\x0b\x32/.google.gapic.metadata.GapicMetadata.MethodList:\x02\x38\x01\x1a\x1d\n\nMethodList\x12\x0f\n\x07methods\x18\x01 \x03(\tB\xba\x01\n\x19\x63om.google.gapic.metadataB\x12GapicMetadataProtoP\x01Z=google.golang.org/genproto/googleapis/gapic/metadata;metadata\xaa\x02\x15Google.Gapic.Metadata\xca\x02\x15Google\\Gapic\\Metadata\xea\x02\x17Google::Gapic::Metadatab\x06proto3'
)


_GAPICMETADATA = DESCRIPTOR.message_types_by_name["GapicMetadata"]
_GAPICMETADATA_SERVICESENTRY = _GAPICMETADATA.nested_types_by_name["ServicesEntry"]
_GAPICMETADATA_SERVICEFORTRANSPORT = _GAPICMETADATA.nested_types_by_name[
    "ServiceForTransport"
]
_GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY = (
    _GAPICMETADATA_SERVICEFORTRANSPORT.nested_types_by_name["ClientsEntry"]
)
_GAPICMETADATA_SERVICEASCLIENT = _GAPICMETADATA.nested_types_by_name["ServiceAsClient"]
_GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY = (
    _GAPICMETADATA_SERVICEASCLIENT.nested_types_by_name["RpcsEntry"]
)
_GAPICMETADATA_METHODLIST = _GAPICMETADATA.nested_types_by_name["MethodList"]
GapicMetadata = _reflection.GeneratedProtocolMessageType(
    "GapicMetadata",
    (_message.Message,),
    {
        "ServicesEntry": _reflection.GeneratedProtocolMessageType(
            "ServicesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _GAPICMETADATA_SERVICESENTRY,
                "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.ServicesEntry)
            },
        ),
        "ServiceForTransport": _reflection.GeneratedProtocolMessageType(
            "ServiceForTransport",
            (_message.Message,),
            {
                "ClientsEntry": _reflection.GeneratedProtocolMessageType(
                    "ClientsEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY,
                        "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                        # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.ServiceForTransport.ClientsEntry)
                    },
                ),
                "DESCRIPTOR": _GAPICMETADATA_SERVICEFORTRANSPORT,
                "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.ServiceForTransport)
            },
        ),
        "ServiceAsClient": _reflection.GeneratedProtocolMessageType(
            "ServiceAsClient",
            (_message.Message,),
            {
                "RpcsEntry": _reflection.GeneratedProtocolMessageType(
                    "RpcsEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY,
                        "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                        # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.ServiceAsClient.RpcsEntry)
                    },
                ),
                "DESCRIPTOR": _GAPICMETADATA_SERVICEASCLIENT,
                "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.ServiceAsClient)
            },
        ),
        "MethodList": _reflection.GeneratedProtocolMessageType(
            "MethodList",
            (_message.Message,),
            {
                "DESCRIPTOR": _GAPICMETADATA_METHODLIST,
                "__module__": "google.gapic.metadata.gapic_metadata_pb2"
                # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata.MethodList)
            },
        ),
        "DESCRIPTOR": _GAPICMETADATA,
        "__module__": "google.gapic.metadata.gapic_metadata_pb2"
        # @@protoc_insertion_point(class_scope:google.gapic.metadata.GapicMetadata)
    },
)
_sym_db.RegisterMessage(GapicMetadata)
_sym_db.RegisterMessage(GapicMetadata.ServicesEntry)
_sym_db.RegisterMessage(GapicMetadata.ServiceForTransport)
_sym_db.RegisterMessage(GapicMetadata.ServiceForTransport.ClientsEntry)
_sym_db.RegisterMessage(GapicMetadata.ServiceAsClient)
_sym_db.RegisterMessage(GapicMetadata.ServiceAsClient.RpcsEntry)
_sym_db.RegisterMessage(GapicMetadata.MethodList)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\031com.google.gapic.metadataB\022GapicMetadataProtoP\001Z=google.golang.org/genproto/googleapis/gapic/metadata;metadata\252\002\025Google.Gapic.Metadata\312\002\025Google\\Gapic\\Metadata\352\002\027Google::Gapic::Metadata"
    _GAPICMETADATA_SERVICESENTRY._options = None
    _GAPICMETADATA_SERVICESENTRY._serialized_options = b"8\001"
    _GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY._options = None
    _GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY._serialized_options = b"8\001"
    _GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY._options = None
    _GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY._serialized_options = b"8\001"
    _GAPICMETADATA._serialized_start = 70
    _GAPICMETADATA._serialized_end = 822
    _GAPICMETADATA_SERVICESENTRY._serialized_start = 256
    _GAPICMETADATA_SERVICESENTRY._serialized_end = 361
    _GAPICMETADATA_SERVICEFORTRANSPORT._serialized_start = 364
    _GAPICMETADATA_SERVICEFORTRANSPORT._serialized_end = 575
    _GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY._serialized_start = 475
    _GAPICMETADATA_SERVICEFORTRANSPORT_CLIENTSENTRY._serialized_end = 575
    _GAPICMETADATA_SERVICEASCLIENT._serialized_start = 578
    _GAPICMETADATA_SERVICEASCLIENT._serialized_end = 791
    _GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY._serialized_start = 699
    _GAPICMETADATA_SERVICEASCLIENT_RPCSENTRY._serialized_end = 791
    _GAPICMETADATA_METHODLIST._serialized_start = 793
    _GAPICMETADATA_METHODLIST._serialized_end = 822
# @@protoc_insertion_point(module_scope)



from abc import (
    ABCMeta,
    abstractmethod,
)

from typing import (
    Dict,
    List,
)

from devicemanager.model import (
    DeviceStatus,
    DeviceStatusHistory,
)


class DeviceManagerRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_disabled_devices(self):
        # type: (...) -> Dict[str, DeviceStatus]

        raise NotImplementedError()

    def get_devices_with_last_history(
        self,
        device_names,  # type: List[str]
    ):
        # type: (...) -> List[DeviceStatus]

        raise NotImplementedError()

    def create_device_status(
        self,
        device_name,  # type: str
    ):
        # type (...) -> None

        raise NotImplementedError()

    def delete_device_status(
        self,
        device_name,  # type: str
    ):
        # type (...) -> None

        raise NotImplementedError()

    def create_device_status_history(
        self,
        device_name,            # type: str
        device_status_history,  # type: DeviceStatusHistory
    ):
        # type (...) -> None

        raise NotImplementedError()



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import re
from collections import namedtuple


def py_type_name(type_name):
    """Get the Python type name for a given model type.

        >>> py_type_name('list')
        'list'
        >>> py_type_name('structure')
        'dict'

    :rtype: string
    """
    return {
        'blob': 'bytes',
        'character': 'string',
        'double': 'float',
        'long': 'integer',
        'map': 'dict',
        'structure': 'dict',
        'timestamp': 'datetime',
    }.get(type_name, type_name)


def py_default(type_name):
    """Get the Python default value for a given model type.

        >>> py_default('string')
        '\'string\''
        >>> py_default('list')
        '[...]'
        >>> py_default('unknown')
        '...'

    :rtype: string
    """
    return {
        'double': '123.0',
        'long': '123',
        'integer': '123',
        'string': "'string'",
        'blob': "b'bytes'",
        'boolean': 'True|False',
        'list': '[...]',
        'map': '{...}',
        'structure': '{...}',
        'timestamp': 'datetime(2015, 1, 1)',
    }.get(type_name, '...')


def get_official_service_name(service_model):
    """Generate the official name of an AWS Service

    :param service_model: The service model representing the service
    """
    official_name = service_model.metadata.get('serviceFullName')
    short_name = service_model.metadata.get('serviceAbbreviation', '')
    if short_name.startswith('Amazon'):
        short_name = short_name[7:]
    if short_name.startswith('AWS'):
        short_name = short_name[4:]
    if short_name and short_name.lower() not in official_name.lower():
        official_name += f' ({short_name})'
    return official_name


_DocumentedShape = namedtuple(
    'DocumentedShape',
    [
        'name',
        'type_name',
        'documentation',
        'metadata',
        'members',
        'required_members',
    ],
)


class DocumentedShape(_DocumentedShape):
    """Use this class to inject new shapes into a model for documentation"""

    def __new__(
        cls,
        name,
        type_name,
        documentation,
        metadata=None,
        members=None,
        required_members=None,
    ):
        if metadata is None:
            metadata = []
        if members is None:
            members = []
        if required_members is None:
            required_members = []
        return super().__new__(
            cls,
            name,
            type_name,
            documentation,
            metadata,
            members,
            required_members,
        )


class AutoPopulatedParam:
    def __init__(self, name, param_description=None):
        self.name = name
        self.param_description = param_description
        if param_description is None:
            self.param_description = (
                'Please note that this parameter is automatically populated '
                'if it is not provided. Including this parameter is not '
                'required\n'
            )

    def document_auto_populated_param(self, event_name, section, **kwargs):
        """Documents auto populated parameters

        It will remove any required marks for the parameter, remove the
        parameter from the example, and add a snippet about the parameter
        being autopopulated in the description.
        """
        if event_name.startswith('docs.request-params'):
            if self.name in section.available_sections:
                section = section.get_section(self.name)
                if 'is-required' in section.available_sections:
                    section.delete_section('is-required')
                description_section = section.get_section(
                    'param-documentation'
                )
                description_section.writeln(self.param_description)
        elif event_name.startswith('docs.request-example'):
            section = section.get_section('structure-value')
            if self.name in section.available_sections:
                section.delete_section(self.name)


class HideParamFromOperations:
    """Hides a single parameter from multiple operations.

    This method will remove a parameter from documentation and from
    examples. This method is typically used for things that are
    automatically populated because a user would be unable to provide
    a value (e.g., a checksum of a serialized XML request body)."""

    def __init__(self, service_name, parameter_name, operation_names):
        """
        :type service_name: str
        :param service_name: Name of the service to modify.

        :type parameter_name: str
        :param parameter_name: Name of the parameter to modify.

        :type operation_names: list
        :param operation_names: Operation names to modify.
        """
        self._parameter_name = parameter_name
        self._params_events = set()
        self._example_events = set()
        # Build up the sets of relevant event names.
        param_template = 'docs.request-params.%s.%s.complete-section'
        example_template = 'docs.request-example.%s.%s.complete-section'
        for name in operation_names:
            self._params_events.add(param_template % (service_name, name))
            self._example_events.add(example_template % (service_name, name))

    def hide_param(self, event_name, section, **kwargs):
        if event_name in self._example_events:
            # Modify the structure value for example events.
            section = section.get_section('structure-value')
        elif event_name not in self._params_events:
            return
        if self._parameter_name in section.available_sections:
            section.delete_section(self._parameter_name)


class AppendParamDocumentation:
    """Appends documentation to a specific parameter"""

    def __init__(self, parameter_name, doc_string):
        self._parameter_name = parameter_name
        self._doc_string = doc_string

    def append_documentation(self, event_name, section, **kwargs):
        if self._parameter_name in section.available_sections:
            section = section.get_section(self._parameter_name)
            description_section = section.get_section('param-documentation')
            description_section.writeln(self._doc_string)


_CONTROLS = {
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
    '\b': '\\b',
    '\f': '\\f',
}
# Combines all CONTROLS keys into a big or regular expression
_ESCAPE_CONTROLS_RE = re.compile('|'.join(map(re.escape, _CONTROLS)))
# Based on the match get the appropriate replacement from CONTROLS
_CONTROLS_MATCH_HANDLER = lambda match: _CONTROLS[match.group(0)]


def escape_controls(value):
    return _ESCAPE_CONTROLS_RE.sub(_CONTROLS_MATCH_HANDLER, value)



import json
import logging

import sysactions

from xml.etree import cElementTree as eTree
from typing import (
    Optional,
    Dict,
    List,
    Union,
    Any,
)

from mw_helper import show_filtered_list_box_dialog, show_numpad_dialog, show_message_dialog
from sysactions import (
    show_messagebox,
    get_current_order,
    StopAction,
    get_model,
    show_keyboard,
    show_any_dialog,
)
from helper import (
    remove_accents,
)

from actions.models import PaymentType
from actions.coupon.apply_coupon import apply_coupon
from actions.discount.get_discount import get_visible_discounts
from actions.products import getProducts
from actions.util import (
    get_authorization,
    get_user_level,
)
from application.domain import SaleLineIdentifier
from manager.reports import justification_report
from model.customexception import DiscountMinLevelException, ErrorApplyingDiscount

logger = logging.getLogger("PosActions")


def apply_discount(
    pos_id,                     # type: str
    selected_coupon=None,       # type: Optional[Dict]
    amount=0.0,                 # type: Optional[float]
    user_auth_level=-1,         # type: Optional[int]
    user_auth_name=None,        # type: Optional[str]
    user_auth_id=None,          # type: Optional[str]
    payment_info=None,          # type: Optional[Dict[str, Any]]
):
    # type: (...) -> bool

    try:
        coupons = get_visible_discounts(pos_id)
        selected_coupon = _select_coupon(pos_id, selected_coupon, coupons)
        if not selected_coupon:
            return False

        coupon = selected_coupon.get("coupon", {})
        model = get_model(pos_id)

        try:
            if coupon.get("requestJustification", False):
                if user_auth_id is None:
                    operator = model.find("Operator")
                    user_auth_id = operator.get("id")

                _add_justification_on_coupon(
                    coupon=coupon,
                    pos_id=pos_id,
                    user_auth_id=user_auth_id,
                )

            selected_sale_lines = _get_selected_sale_lines(
                coupon=coupon,
                model=model,
                pos_id=pos_id,
            )

            if coupon.get("askValue", False) and not amount:
                coupon["index"] = _get_discount_type(coupon)
                amount = amount or _get_discount_amount(coupon, pos_id)
                coupon = set_coupon_amount(
                    coupon=coupon,
                    amount=amount,
                )

            benefit_applied = _add_benefit(
                pos_id=pos_id,
                coupon=coupon,
                user_auth_level=user_auth_level,
                sale_lines=selected_sale_lines,
                payment_info=payment_info,
            )

            if benefit_applied:
                _check_and_print_justification(
                    pos_id=pos_id,
                    user_auth=user_auth_name,
                    coupon=coupon,
                )

            return benefit_applied
        except DiscountMinLevelException as ex:
            show_messagebox(pos_id, "$NDISCOUNT_ERROR_USER_MIN_LEVEL")
            user_was_authorized, user_auth = get_authorization(
                pos_id=pos_id,
                min_level=ex.min_level_required,
                model=model
            )
            if user_was_authorized:
                user_xml_str = sysactions.get_user_information(user_auth)
                user_element = eTree.XML(user_xml_str).find("user")
                user_auth_level = int(user_element.attrib["Level"])
                user_auth_id = str(user_element.attrib["UserId"])
                coupon['authorizedBy'] = user_auth_id
                selected_coupon["coupon"] = coupon

                apply_discount(
                    pos_id=pos_id,
                    selected_coupon=selected_coupon,
                    amount=amount,
                    user_auth_level=user_auth_level,
                    user_auth_name=user_element.attrib["UserName"],
                    user_auth_id=user_auth_id,
                )
                return True

    except ErrorApplyingDiscount as ex:
        logger.warn("Error applying discount ${}".format(ex.error_code))
        show_message_dialog(pos_id, "$ERROR", "${}".format(ex.error_code))
    except Exception as ex:
        logger.exception("Error applying discount")
        message = "FATAL_ERROR"
        if hasattr(ex, 'message'):
            message = "${}".format(message)
        show_message_dialog(pos_id, "$ERROR", message)

    return False


def _select_coupon(
    pos_id,             # type: str
    selected_coupon,    # type: Optional[Dict]
    coupons,            # type: List
):
    # type: (...) -> Optional[Dict]

    if selected_coupon:
        return selected_coupon

    if len(coupons) == 0:
        show_messagebox(pos_id, "$DISCOUNT_NOT_FOUND", icon="warning")
        return None

    selected_index = _select_discount(pos_id, coupons)
    if selected_index is None:
        return None

    selected_coupon = filter(lambda c: c["index"] == selected_index + 1, coupons)
    if not selected_coupon:
        return None

    return selected_coupon[0]


def _get_discount_amount(coupon, pos_id):
    model = get_model(pos_id)
    order = get_current_order(model)
    amount, percentage = _ask_discount_amount(pos_id, order, coupon=coupon)
    if percentage is not None:
        amount = percentage

    return amount


def apply_amount_discount(pos_id, pos_ot, order, coupon=None, discount_type=None, discount_value=None):
    amount, _ = _ask_discount_amount(pos_id, order, coupon, discount_type, discount_value)
    pos_ot.applyDiscount(1, amount, forcebyitem=1)


def _ask_discount_amount(pos_id, order, coupon=None, discount_type=None, discount_value=None):
    if not discount_type:
        discount_type = coupon['index']
        title = "$DISCOUNT_VALUE_TITLE" if discount_type == PaymentType.AMOUNT.value else "$DISCOUNT_PERCENTAGE_TITLE"
        mask = "CURRENCY" if discount_type == PaymentType.AMOUNT.value else "NUMBER"
        if not discount_value:
            discount_value = show_numpad_dialog(pos_id, title, "", mask)
            if discount_value in [None, '']:
                raise StopAction()

    if discount_type == PaymentType.AMOUNT.value:
        return discount_value, None

    amount = float(order.get("totalGross"))

    return (amount * discount_value) / 100, discount_value


def _select_discount(pos_id, coupons):
    options = ["{0} - {1}".format(coupon["index"], coupon["name"].encode("UTF-8")) for coupon in coupons]
    return show_filtered_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", mask="NOFILTER")


def _add_benefit(
    pos_id,             # type: str
    coupon,             # type: Dict[str, Any]
    user_auth_level,    # type: int
    sale_lines=None,    # type: Optional[List[SaleLineIdentifier]]
    payment_info=None,  # type: Optional[Dict[str, Any]]
):
    # type: (...) -> bool

    authorization_level = int(coupon.get("conditions", {}).get("authorizationLevel", 0))
    if _get_discount_authorization(
        pos_id=pos_id,
        coupon=coupon,
        authorization_level=authorization_level,
    ) is False:
        raise StopAction()

    ret = apply_coupon(
        pos_id=pos_id,
        table_id=None,
        user_auth_level=user_auth_level,
        benefit=coupon,
        sale_lines=sale_lines,
        payment_info=payment_info,
    )

    return ret


def _get_object_json(json_object, key, value):
    return [obj for obj in json_object if obj[key] == value][0]


def set_coupon_amount(coupon, amount):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        value_actions = ["discountAmountOnTotal", "discountPercentageOnTotal"]
        group_actions = ["discountPercentageByGroupsOnTotal", "discountAmountByGroupsOnTotal"]
        if any(x for x in value_actions if x in action.keys()):
            action_name = action.keys()[0]
            action[action_name] = amount
            return coupon
        elif any(x for x in group_actions if x in action.keys()):
            action.values()[0][0]["amount"] = amount
            return coupon

    raise StopAction()


def _get_discount_type(coupon):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        amount_actions = ["discountAmountOnTotal", "discountAmountByGroupsOnTotal"]
        percentage_actions = ["discountPercentageOnTotal", "discountPercentageByGroupsOnTotal"]
        if any(x for x in amount_actions if x in action.keys()):
            return PaymentType.AMOUNT.value
        elif any(x for x in percentage_actions if x in action.keys()):
            return PaymentType.PERCENTAGE.value

    raise StopAction()


def _get_justification_discount(
    pos_id,     # type: str
):
    # type: (...) -> str

    while True:
        justification = show_keyboard(
            posid=pos_id,
            message="$JUSTIFICATION_DISCOUNT",
            title='',
            timeout=1800000,
            buttons="$OK",
        )
        if justification is None:
            raise StopAction()

        justification = justification.strip()
        if len(justification) < 5:
            show_messagebox(
                posid=pos_id,
                message="$MINIMUM_5_CHARACTERS_REQUIRED",
            )
            continue

        break

    return justification


def _get_discount_authorization(
    pos_id,                 # type: str
    coupon,                 # type: dict
    authorization_level,    # type: int
):
    # type: (...) -> Optional[dict, bool]

    model = get_model(posid=pos_id)
    operator = model.find("./Operator")
    user_id = operator.get("id")
    current_user_level = get_user_level(user_id=user_id)

    if current_user_level >= authorization_level:
        coupon['authorized_by'] = user_id
        return coupon

    user_was_authorized, user_auth = get_authorization(
        pos_id=pos_id,
        min_level=authorization_level,
        model=model,
        access_denied_message="$NDISCOUNT_ERROR_USER_MIN_LEVEL",
    )
    authorized_user_level = get_user_level(user_id=user_auth)
    if user_was_authorized and authorized_user_level >= authorization_level:
        coupon['authorized_by'] = user_auth
        return coupon

    return False


def _check_and_print_justification(
    pos_id,         # type: str
    user_auth,      # type: str
    coupon,         # type: Dict
):

    justification = coupon.get("justification")
    should_print_justification = coupon.get("printJustifyCoupon", False)
    if not justification or not should_print_justification:
        return

    justification_report(
        pos_id=pos_id,
        justification=justification,
        discount_description=coupon.get('userDescription', ""),
        authorized_by=user_auth,
    )


def _add_justification_on_coupon(
    coupon,         # type: Dict[str, Any]
    pos_id,         # type: str
    user_auth_id,   # type: str
):
    # type: (...) -> None

    justification = coupon.get("justification")
    if justification is None:
        justification = _get_justification_discount(
            pos_id=pos_id,
        )
        coupon['justification'] = justification

    coupon['appliedBy'] = user_auth_id


def _get_selected_sale_lines(
    coupon,  # type: Dict[str, Any]
    model,   # type: eTree.Element
    pos_id,  # type: str
):
    # type: (...) -> List[SaleLineIdentifier]

    if not coupon.get("conditions", {}).get("askItemToApplyDiscount"):
        return []

    order = get_current_order(model=model)
    options = _get_options_to_apply_discount(
        pos_id=pos_id,
        sale_lines=order.findall("SaleLine"),
        discount_details=_get_discount_details(order),
        coupon=coupon,
    )

    if not options:
        show_messagebox(pos_id, "$NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE")
        raise StopAction()

    selected_items = _show_multiple_select_list_box_dialog(
        pos_id=pos_id,
        options=map(json.dumps, options),
        title="$SELECT_ONE_OR_MORE_ITEMS",
    )

    if not selected_items:
        raise StopAction()

    selected_sale_lines = [
        SaleLineIdentifier(
            line_number=int(options[i].get("lineNumber")),
            item_id=options[i].get("context"),
            level=int(options[i].get("level")),
            part_code=int(options[i].get("partCode")),
        )
        for i in selected_items
    ]

    return selected_sale_lines


def _get_options_to_apply_discount(
    pos_id,             # type: str
    sale_lines,         # type: eTree.Element
    discount_details,   # type: Dict[str, float]
    coupon,             # type: Dict[str, Any]
):
    # type: (...) -> List[Dict[str, Any]]

    options = {}
    item_level = {}
    line_number_status = {}

    groups_and_items_restrictions = _get_action_groups_and_items_restrictions(coupon)
    all_products = json.loads(getProducts(pos_id=pos_id))

    for sale_line in sale_lines:
        line_number = sale_line.get("lineNumber")
        context = sale_line.get("itemId")
        part_code = sale_line.get("partCode")
        item_price = float(sale_line.get("itemPrice", 0))
        product_name = sale_line.get("productName")
        parent_key = "{}.{}".format(
            line_number,
            context,
        )
        my_key = parent_key + "." + part_code

        not_cfh = all_products.get(part_code, {}).get("CFH") != "true"
        qty = float(sale_line.get("qty", 0))
        if qty > 0 and not_cfh and sale_line.get("itemType") != "OPTION":
            level = sale_line.get("level")
            sale_line_identifier = "{}.{}-{}.{}".format(
                line_number,
                level,
                context,
                part_code,
            )
            item_total = item_price - discount_details.get(sale_line_identifier, 0)
            option = {
                "lineNumber": line_number,
                "level": level,
                "context": context,
                "partCode": part_code,
                "qty": qty,
                "productName": remove_accents(text=product_name),
                "itemPrice": item_price,
                "itemTotal": "{:.3f}".format(item_total),
                "status": _get_sale_line_status(
                    groups_and_items_restrictions=groups_and_items_restrictions,
                    sale_line=sale_line,
                    item_total=item_total,
                )
            }

            if line_number not in options:
                options[line_number] = []
                line_number_status[line_number] = False

            options[line_number].append(option)
            line_number_status[line_number] |= option.get("status") == "enabled"

            item_level[my_key] = item_level.get(parent_key, -1) + 1
        else:
            item_level[my_key] = item_level.get(parent_key, -1)

    filtered_options = _filter_line_number_all_disabled(
        line_number_status=line_number_status,
        options=options,
    )

    return filtered_options


def _filter_line_number_all_disabled(
    line_number_status,     # type: Dict
    options,                # type: Dict
):
    # type: (...) -> List

    filtered_options = []
    for level, values in options.items():
        if line_number_status[level]:
            filtered_options.extend(values)

    return filtered_options


def _get_sale_line_status(
    groups_and_items_restrictions,      # type: Dict[str, Any]
    sale_line,                          # type: eTree.Element
    item_total,                         # type: float
):
    # type: (...) -> str

    if item_total == 0:
        return "disabled"

    groups = groups_and_items_restrictions["groups"]
    items = groups_and_items_restrictions["items"]

    if not groups and not items:
        return "enabled"

    if groups:
        sale_line_groups = json.loads(sale_line.get("jsonArrayTags", "[]"))
        valid_group = any(["PLUFamilyGroup={}".format(group) in sale_line_groups for group in groups])
        if valid_group:
            return "enabled"

    if sale_line.get("partCode") in items:
        return "enabled"

    for part_code in sale_line.get("itemId").split(".")[1:]:
        if part_code in items:
            return "enabled"

    return "disabled"


def _get_discount_details(
    order,  # type: eTree.Element
):
    # type: (...) -> Dict[str, float]

    benefit_list = order.find(".//CustomOrderProperties/OrderProperty[@key='BENEFIT_LIST']")
    benefit_list = json.loads(benefit_list.get("value") or "{}") if benefit_list is not None else {}
    discount_values = benefit_list.get("discountItemValues", {})

    discount_details = {}
    if not discount_values:
        return discount_details

    for _, value in discount_values.items():
        for sale_line_key, amount in value.items():
            if sale_line_key in discount_details:
                discount_details[sale_line_key] += amount
            else:
                discount_details[sale_line_key] = amount

    return discount_details


def _get_action_groups_and_items_restrictions(
    coupon,     # type: Dict[str, Any]
):
    # type: (...) -> Dict[str, List]

    groups_and_items_restrictions = {
        "groups": [],
        "items": [],
    }
    if not coupon.get("actions") or not coupon.get("actions").get("0"):
        return groups_and_items_restrictions

    items_actions = [
        "discountPercentageByItems",
        "discountAmountByItems",
        "discountPercentageByItemsOnTotal",
        "discountAmountByItemsOnTotal",
    ]
    groups_and_items_restrictions["items"].extend(
        _populate_groups_and_items_restrictions(
            coupon=coupon,
            actions=items_actions,
            event_key="item",
        )
    )

    group_actions = [
        "discountPercentageByGroups",
        "discountAmountByGroups",
        "discountPercentageByGroupsOnTotal",
        "discountAmountByGroupsOnTotal",
    ]
    groups_and_items_restrictions["groups"].extend(
        _populate_groups_and_items_restrictions(
            coupon=coupon,
            actions=group_actions,
            event_key="group",
        )
    )

    return groups_and_items_restrictions


def _populate_groups_and_items_restrictions(
    coupon,         # type: Dict[str, Any]
    actions,        # type: List[str]
    event_key,      # type: str
):
    # type: (...) -> List

    action = coupon.get("actions").get("0").items()[0]
    action_name = action[0]
    if action_name not in actions:
        return []

    restictions = []
    action_events = action[1]
    for event in action_events:
        restictions.append(event.get(event_key))

    return restictions


def _show_multiple_select_list_box_dialog(
    pos_id,     # type: str
    options,    # type: Union[str, List[str]]
    title="",   # type: str
):
    # type: (...) -> Optional[List[int]]

    if isinstance(options, list):
        options = "|".join(options)

    response = show_any_dialog(
        posid=pos_id,
        dlgtype="multipleSelectListBox",
        message="",
        title=title,
        infotext=options,
        buttons="$OK|$CANCEL",
        icon="",
        timeout=600000,
        mask="",
        defvalue="",
        minvalue="",
        maxvalue="",
        contents=None,
        asynch=False,
    )

    if response is None:
        return None

    return map(int, response.split(","))



from production.manager import ProductionManager
from production.command import DevicesUpdatedCommand


class DeviceStatusChangedInteractor(object):
    def __init__(
        self,
        production_manager,     # type: ProductionManager
    ):
        # type: (...) -> None

        self.production_manager = production_manager

    def execute(
        self,   # type: DeviceStatusChangedInteractor
    ):
        # type: (...) -> None

        self.production_manager.handle_command(command=DevicesUpdatedCommand())



from application.domain import TableService  # noqa
from application.interactor.service import TableReleaser


class TableReleaserImpl(TableReleaser):
    def __init__(self, table_service):
        # type: (TableService) -> None
        self.table_service = table_service

    def release_all(self, pos_id):
        tables = self.table_service.list_tables(pos_id)
        for table in tables:
            if pos_id == table.current_pos_id:
                self.table_service.store_service(pos_id, table.id)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

from enum import Enum

import six

from cryptography import utils
from cryptography.x509.oid import NameOID, ObjectIdentifier


class _ASN1Type(Enum):
    UTF8String = 12
    NumericString = 18
    PrintableString = 19
    T61String = 20
    IA5String = 22
    UTCTime = 23
    GeneralizedTime = 24
    VisibleString = 26
    UniversalString = 28
    BMPString = 30


_ASN1_TYPE_TO_ENUM = dict((i.value, i) for i in _ASN1Type)
_SENTINEL = object()
_NAMEOID_DEFAULT_TYPE = {
    NameOID.COUNTRY_NAME: _ASN1Type.PrintableString,
    NameOID.JURISDICTION_COUNTRY_NAME: _ASN1Type.PrintableString,
    NameOID.SERIAL_NUMBER: _ASN1Type.PrintableString,
    NameOID.DN_QUALIFIER: _ASN1Type.PrintableString,
    NameOID.EMAIL_ADDRESS: _ASN1Type.IA5String,
    NameOID.DOMAIN_COMPONENT: _ASN1Type.IA5String,
}

#: Short attribute names from RFC 4514:
#: https://tools.ietf.org/html/rfc4514#page-7
_NAMEOID_TO_NAME = {
    NameOID.COMMON_NAME: 'CN',
    NameOID.LOCALITY_NAME: 'L',
    NameOID.STATE_OR_PROVINCE_NAME: 'ST',
    NameOID.ORGANIZATION_NAME: 'O',
    NameOID.ORGANIZATIONAL_UNIT_NAME: 'OU',
    NameOID.COUNTRY_NAME: 'C',
    NameOID.STREET_ADDRESS: 'STREET',
    NameOID.DOMAIN_COMPONENT: 'DC',
    NameOID.USER_ID: 'UID',
}


def _escape_dn_value(val):
    """Escape special characters in RFC4514 Distinguished Name value."""

    # See https://tools.ietf.org/html/rfc4514#section-2.4
    val = val.replace('\\', '\\\\')
    val = val.replace('"', '\\"')
    val = val.replace('+', '\\+')
    val = val.replace(',', '\\,')
    val = val.replace(';', '\\;')
    val = val.replace('<', '\\<')
    val = val.replace('>', '\\>')
    val = val.replace('\0', '\\00')

    if val[0] in ('#', ' '):
        val = '\\' + val
    if val[-1] == ' ':
        val = val[:-1] + '\\ '

    return val


class NameAttribute(object):
    def __init__(self, oid, value, _type=_SENTINEL):
        if not isinstance(oid, ObjectIdentifier):
            raise TypeError(
                "oid argument must be an ObjectIdentifier instance."
            )

        if not isinstance(value, six.text_type):
            raise TypeError(
                "value argument must be a text type."
            )

        if (
            oid == NameOID.COUNTRY_NAME or
            oid == NameOID.JURISDICTION_COUNTRY_NAME
        ):
            if len(value.encode("utf8")) != 2:
                raise ValueError(
                    "Country name must be a 2 character country code"
                )

        if len(value) == 0:
            raise ValueError("Value cannot be an empty string")

        # The appropriate ASN1 string type varies by OID and is defined across
        # multiple RFCs including 2459, 3280, and 5280. In general UTF8String
        # is preferred (2459), but 3280 and 5280 specify several OIDs with
        # alternate types. This means when we see the sentinel value we need
        # to look up whether the OID has a non-UTF8 type. If it does, set it
        # to that. Otherwise, UTF8!
        if _type == _SENTINEL:
            _type = _NAMEOID_DEFAULT_TYPE.get(oid, _ASN1Type.UTF8String)

        if not isinstance(_type, _ASN1Type):
            raise TypeError("_type must be from the _ASN1Type enum")

        self._oid = oid
        self._value = value
        self._type = _type

    oid = utils.read_only_property("_oid")
    value = utils.read_only_property("_value")

    def rfc4514_string(self):
        """
        Format as RFC4514 Distinguished Name string.

        Use short attribute name if available, otherwise fall back to OID
        dotted string.
        """
        key = _NAMEOID_TO_NAME.get(self.oid, self.oid.dotted_string)
        return '%s=%s' % (key, _escape_dn_value(self.value))

    def __eq__(self, other):
        if not isinstance(other, NameAttribute):
            return NotImplemented

        return (
            self.oid == other.oid and
            self.value == other.value
        )

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.oid, self.value))

    def __repr__(self):
        return "<NameAttribute(oid={0.oid}, value={0.value!r})>".format(self)


class RelativeDistinguishedName(object):
    def __init__(self, attributes):
        attributes = list(attributes)
        if not attributes:
            raise ValueError("a relative distinguished name cannot be empty")
        if not all(isinstance(x, NameAttribute) for x in attributes):
            raise TypeError("attributes must be an iterable of NameAttribute")

        # Keep list and frozenset to preserve attribute order where it matters
        self._attributes = attributes
        self._attribute_set = frozenset(attributes)

        if len(self._attribute_set) != len(attributes):
            raise ValueError("duplicate attributes are not allowed")

    def get_attributes_for_oid(self, oid):
        return [i for i in self if i.oid == oid]

    def rfc4514_string(self):
        """
        Format as RFC4514 Distinguished Name string.

        Within each RDN, attributes are joined by '+', although that is rarely
        used in certificates.
        """
        return '+'.join(attr.rfc4514_string() for attr in self._attributes)

    def __eq__(self, other):
        if not isinstance(other, RelativeDistinguishedName):
            return NotImplemented

        return self._attribute_set == other._attribute_set

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._attribute_set)

    def __iter__(self):
        return iter(self._attributes)

    def __len__(self):
        return len(self._attributes)

    def __repr__(self):
        return "<RelativeDistinguishedName({})>".format(self.rfc4514_string())


class Name(object):
    def __init__(self, attributes):
        attributes = list(attributes)
        if all(isinstance(x, NameAttribute) for x in attributes):
            self._attributes = [
                RelativeDistinguishedName([x]) for x in attributes
            ]
        elif all(isinstance(x, RelativeDistinguishedName) for x in attributes):
            self._attributes = attributes
        else:
            raise TypeError(
                "attributes must be a list of NameAttribute"
                " or a list RelativeDistinguishedName"
            )

    def rfc4514_string(self):
        """
        Format as RFC4514 Distinguished Name string.
        For example 'CN=foobar.com,O=Foo Corp,C=US'

        An X.509 name is a two-level structure: a list of sets of attributes.
        Each list element is separated by ',' and within each list element, set
        elements are separated by '+'. The latter is almost never used in
        real world certificates.
        """
        return ','.join(attr.rfc4514_string() for attr in self._attributes)

    def get_attributes_for_oid(self, oid):
        return [i for i in self if i.oid == oid]

    @property
    def rdns(self):
        return self._attributes

    def public_bytes(self, backend):
        return backend.x509_name_bytes(self)

    def __eq__(self, other):
        if not isinstance(other, Name):
            return NotImplemented

        return self._attributes == other._attributes

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        # TODO: this is relatively expensive, if this looks like a bottleneck
        # for you, consider optimizing!
        return hash(tuple(self._attributes))

    def __iter__(self):
        for rdn in self._attributes:
            for ava in rdn:
                yield ava

    def __len__(self):
        return sum(len(rdn) for rdn in self._attributes)

    def __repr__(self):
        if six.PY2:
            return "<Name({})>".format(self.rfc4514_string().encode('utf8'))
        else:
            return "<Name({})>".format(self.rfc4514_string())



"""Read resources contained within a package."""

from ._common import (
    as_file,
    files,
    Package,
)

from .abc import ResourceReader


__all__ = [
    'Package',
    'ResourceReader',
    'as_file',
    'files',
]
